import re
from typing import Any


class XMLParser:
    def __init__(self, xml_string: str):
        self.xml_string = xml_string.strip()
        self.index = 0

    def parse(self) -> dict[str, Any]:
        """Parse the XML string and return the root element as a dictionary."""
        return self._parse_element()

    def _parse_element(self) -> dict[str, Any]:
        """Parse a single XML element and its children."""
        self._skip_whitespace()
        if self.xml_string[self.index] != "<":
            raise ValueError(f"Expected '<', found '{self.xml_string[self.index]}'")

        tag, attributes = self._parse_opening_tag()
        children: dict[str, str | list | dict] = {}
        text = ""

        while self.index < len(self.xml_string):
            self._skip_whitespace()
            if self.xml_string.startswith("</", self.index):
                closing_tag = self._parse_closing_tag()
                if closing_tag != tag:
                    raise ValueError(f"Mismatched tags: '{tag}' and '{closing_tag}'")
                break
            elif self.xml_string.startswith("<", self.index):
                child = self._parse_element()
                child_tag, child_data = next(iter(child.items()))
                if child_tag in children:
                    if not isinstance(children[child_tag], list):
                        children[child_tag] = [children[child_tag]]
                    children[child_tag].append(child_data)
                else:
                    children[child_tag] = child_data
            else:
                text += self._parse_text()

        result: dict[str, Any] = {}
        if attributes:
            result["@attributes"] = attributes
        if children:
            result.update(children)
        elif text.strip():
            result = text.strip()

        return {tag: result}

    def _parse_opening_tag(self) -> tuple[str, dict[str, str]]:
        """Parse an opening XML tag and its attributes."""
        match = re.match(
            r'<(\w+)((?:\s+\w+="[^"]*")*)\s*/?>',
            self.xml_string[self.index :],  # noqa
        )
        if not match:
            raise ValueError("Invalid opening tag")
        self.index += match.end()
        tag = match.group(1)
        attributes = dict(re.findall(r'(\w+)="([^"]*)"', match.group(2)))
        return tag, attributes

    def _parse_closing_tag(self) -> str:
        """Parse a closing XML tag."""
        match = re.match(r"</(\w+)>", self.xml_string[self.index :])  # noqa
        if not match:
            raise ValueError("Invalid closing tag")
        self.index += match.end()
        return match.group(1)

    def _parse_text(self) -> str:
        """Parse text content between XML tags."""
        start = self.index
        while self.index < len(self.xml_string) and self.xml_string[self.index] != "<":
            self.index += 1
        return self.xml_string[start : self.index]  # noqa

    def _skip_whitespace(self) -> None:
        """Skip any whitespace characters at the current parsing position."""
        p_ = len(self.xml_string[self.index :])  # noqa
        m_ = len(self.xml_string[self.index :].lstrip())  # noqa

        self.index += p_ - m_



from typing import Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, model_validator


class GroqEndpointRequestBody(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        use_enum_values=True,
        validate_assignment=True,
        arbitrary_types_allowed=True,
    )


class GroqEndpointResponseBody(BaseModel):
    model_config = ConfigDict(
        use_enum_values=True,
        validate_assignment=True,
        arbitrary_types_allowed=True,
    )


class GroqEndpointQueryParam(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        use_enum_values=True,
        validate_assignment=True,
        arbitrary_types_allowed=True,
    )


class GroqEndpointPathParam(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        use_enum_values=True,
        validate_assignment=True,
        arbitrary_types_allowed=True,
    )


class GroqChatMessage(BaseModel):
    role: str = Field(description="The role of the message sender")
    content: str = Field(description="The content of the message")
    name: str | None = Field(
        default=None, description="Optional name for the message sender"
    )


class GroqChatCompletionRequest(GroqEndpointRequestBody):
    messages: list[GroqChatMessage] = Field(
        description="List of messages in the conversation"
    )
    model: str = Field(description="ID of the model to use")
    temperature: float | None = Field(
        default=1.0,
        description="Sampling temperature between 0 and 2",
        ge=0.0,
        le=2.0,
    )
    top_p: float | None = Field(
        default=1.0,
        description="Nucleus sampling threshold",
        ge=0.0,
        le=1.0,
    )
    n: int | None = Field(
        default=1,
        description="Number of completions to generate",
        ge=1,
    )
    stream: bool | None = Field(
        default=False,
        description="Whether to stream the response",
    )
    stop: str | list[str] | None = Field(
        default=None,
        description="Sequences where the API will stop generating",
    )
    max_tokens: int | None = Field(
        default=None,
        description="Maximum number of tokens to generate",
        ge=1,
    )
    presence_penalty: float | None = Field(
        default=0.0,
        description="Presence penalty",
        ge=-2.0,
        le=2.0,
    )
    frequency_penalty: float | None = Field(
        default=0.0,
        description="Frequency penalty",
        ge=-2.0,
        le=2.0,
    )
    logit_bias: dict[str, float] | None = Field(
        default=None,
        description="Modify token likelihoods",
    )
    user: str | None = Field(
        default=None,
        description="Unique identifier for the end user",
    )

    @model_validator(mode="after")
    def validate_messages(self):
        if not self.messages:
            raise ValueError("Messages list cannot be empty")

        # Validate roles
        valid_roles = {"system", "user", "assistant"}
        for msg in self.messages:
            if msg.role not in valid_roles:
                raise ValueError(
                    f"Invalid role: {msg.role}. Must be one of {valid_roles}"
                )

        return self


class GroqAudioRequest(GroqEndpointRequestBody):
    model: str = Field(description="ID of the model to use")
    file: str = Field(description="Audio file to process")
    language: str | None = Field(
        default=None,
        description="Language of the audio (ISO-639-1)",
    )
    prompt: str | None = Field(
        default=None,
        description="Text to guide the model",
    )
    response_format: str | None = Field(
        default="json",
        description="Response format (json, text, or verbose_json)",
    )
    temperature: float | None = Field(
        default=0,
        description="Sampling temperature",
        ge=0.0,
        le=1.0,
    )
    timestamp_granularities: list[str] | None = Field(
        default=["segment"],
        description="Timestamp granularities (word or segment)",
    )


class GroqChatCompletionResponse(GroqEndpointResponseBody):
    id: str = Field(description="Unique identifier for the completion")
    object: str = Field(description="Object type")
    created: int = Field(description="Unix timestamp of creation")
    model: str = Field(description="Model used")
    choices: list[dict] = Field(description="Generated completions")
    usage: dict = Field(description="Token usage information")
    system_fingerprint: str | None = Field(
        description="System fingerprint for deterministic results"
    )


class GroqAudioResponse(GroqEndpointResponseBody):
    text: str = Field(description="Transcribed or translated text")
    x_groq: dict = Field(description="Groq-specific metadata")


# Request model mapping for iModel
GROQ_REQUEST_MODELS = {
    "create_chat_completion": {
        "request_body": GroqChatCompletionRequest,
    },
    "create_transcription": {
        "request_body": GroqAudioRequest,
    },
    "create_translation": {
        "request_body": GroqAudioRequest,
    },
    "chat/completions": {
        "request_body": GroqChatCompletionRequest,
    },
    "audio/transcriptions": {
        "request_body": GroqAudioRequest,
    },
    "audio/translations": {
        "request_body": GroqAudioRequest,
    },
}



from typing import Literal

from pydantic import BaseModel, Field


class Message(BaseModel):
    role: Literal["user", "assistant"] = Field(
        description="The role of the message sender"
    )
    content: str = Field(description="The content of the message")



from enum import Enum
from typing import List, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field

from ..data_models import OpenAIEndpointResponseBody


class Status(str, Enum):
    VALIDATING_FILES = "validating_files"
    QUEUED = "queued"
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    CANCELLED = "cancelled"


class Error(BaseModel):
    code: str | None = Field(None, description="A machine-readable error code.")

    message: str | None = Field(None, description="A human-readable error message.")

    param: str | None = Field(
        None, description="The parameter that was invalid, if applicable."
    )


class Hyperparameters(BaseModel):
    n_epochs: Literal["auto"] | int = Field(
        description="The number of epochs to train the model for. 'auto' or 1-50.",
    )

    # TODO: check hyperparameters
    batch_size: Literal["auto"] | int = Field(None, description="The batch size.")

    learning_rate_multiplier: Literal["auto"] | int = Field(
        None, description="The learning rate multiplier."
    )

    model_config = ConfigDict(extra="allow")


class Wandb(BaseModel):
    project: str = Field(..., description="The name of the project for the new run.")

    name: str | None = Field(None, description="A display name for the run.")

    entity: str | None = Field(None, description="The entity to use for the run.")

    tags: list[str] | None = Field(description="A list of tags for the run.")


class Integration(BaseModel):
    type: str = Field(..., description="The type of the integration being enabled.")

    wandb: Wandb | None = Field(
        None, description="Settings for Weights and Biases integration."
    )


class OpenAIFineTuningJobResponseBody(OpenAIEndpointResponseBody):
    id: str = Field(
        description="The object identifier, referenceable in API endpoints."
    )

    created_at: int = Field(
        ..., description="The Unix timestamp (in seconds) of job creation."
    )

    error: Error | None = Field(None, description="Error information for failed jobs.")

    fine_tuned_model: str | None = Field(
        None, description="The name of the fine-tuned model being created."
    )

    finished_at: int | None = Field(
        None, description="The Unix timestamp (in seconds) of job completion."
    )

    hyperparameters: Hyperparameters = Field(
        description="The hyperparameters used for the fine-tuning job."
    )

    model: str = Field(description="The base model being fine-tuned.")

    object: Literal["fine_tuning.job"] = Field(
        description="The object type, always 'fine_tuning.job'."
    )

    organization_id: str = Field(
        ..., description="The organization that owns the fine-tuning job."
    )

    result_files: list[str] = Field(
        ..., description="The compiled results file ID(s) for the job."
    )

    status: Status = Field(
        description=(
            "The current status of the fine-tuning job: validating_files, "
            "queued, running, succeeded, failed, or cancelled."
        ),
    )

    trained_tokens: int | None = Field(
        None, description="The total number of billable tokens processed."
    )

    training_file: str = Field(description="The file ID used for training.")

    validation_file: str | None = Field(
        None, description="The file ID used for validation."
    )

    integrations: list[Integration] | None = Field(
        None, description="A list of integrations enabled for this job."
    )

    seed: int = Field(None, description="The seed used for the fine-tuning job.")

    estimated_finish: int | None = Field(
        None,
        description="The Unix timestamp (in seconds) for estimated job completion.",
    )



from typing import List

from pydantic import BaseModel, Field

from ..data_models import OpenAIEndpointResponseBody


class Categories(BaseModel):
    hate: bool = Field(
        description="Content that expresses, incites, or promotes hate based on race, gender, ethnicity, "
        "religion, nationality, sexual orientation, disability status, or caste. Hateful "
        "content aimed at non-protected groups (e.g., chess players) is harassment."
    )

    hate_threatening: bool = Field(
        alias="hate/threatening",
        description="Hateful content that also includes violence or serious harm towards the targeted "
        "group based on race, gender, ethnicity, religion, nationality, sexual orientation, "
        "disability status, or caste.",
    )

    harassment: bool = Field(
        description="Content that expresses, incites, or promotes harassing language towards any target."
    )

    harassment_threatening: bool = Field(
        alias="harassment/threatening",
        description="Harassment content that also includes violence or serious harm towards any target.",
    )

    self_harm: bool = Field(
        alias="self-harm",
        description="Content that promotes, encourages, or depicts acts of self-harm, such as suicide, "
        "cutting, and eating disorders.",
    )

    self_harm_intent: bool = Field(
        alias="self-harm/intent",
        description="Content where the speaker expresses that they are engaging or intend to engage in "
        "acts of self-harm, such as suicide, cutting, and eating disorders.",
    )

    self_harm_instructions: bool = Field(
        alias="self-harm/instructions",
        description="Content that encourages performing acts of self-harm, such as suicide, cutting, "
        "and eating disorders, or that gives instructions or advice on how to commit such acts.",
    )

    sexual: bool = Field(
        description="Content meant to arouse sexual excitement, such as the description of sexual activity, "
        "or that promotes sexual services (excluding sex education and wellness)."
    )

    sexual_minors: bool = Field(
        alias="sexual/minors",
        description="Sexual content that includes an individual who is under 18 years old.",
    )

    violence: bool = Field(
        description="Content that depicts death, violence, or physical injury."
    )

    violence_graphic: bool = Field(
        alias="violence/graphic",
        description="Content that depicts death, violence, or physical injury in graphic detail.",
    )


class CategoryScores(BaseModel):
    hate: float = Field(description="The score for the category 'hate'.")

    hate_threatening: float = Field(
        alias="hate/threatening",
        description="The score for the category 'hate/threatening'.",
    )

    harassment: float = Field(description="The score for the category 'harassment'.")

    harassment_threatening: float = Field(
        alias="harassment/threatening",
        description="The score for the category 'harassment/threatening'.",
    )

    self_harm: float = Field(
        alias="self-harm",
        description="The score for the category 'self-harm'.",
    )

    self_harm_intent: float = Field(
        alias="self-harm/intent",
        description="The score for the category 'self-harm/intent'.",
    )

    self_harm_instructions: float = Field(
        alias="self-harm/instructions",
        description="The score for the category 'self-harm/instructions'.",
    )

    sexual: float = Field(description="The score for the category 'sexual'.")

    sexual_minors: float = Field(
        alias="sexual/minors",
        description="The score for the category 'sexual/minors'.",
    )

    violence: float = Field(description="The score for the category 'violence'.")

    violence_graphic: float = Field(
        alias="violence/graphic",
        description="The score for the category 'violence/graphic'.",
    )


class Result(BaseModel):
    flagged: bool = Field(
        description="Whether the content violates OpenAI's usage policies."
    )

    categories: Categories = Field(
        description="A dictionary of per-category binary content policy violation flags.",
    )

    category_scores: CategoryScores = Field(
        description="A dictionary of per-category raw scores output by the model."
    )


class OpenAIModerationResponseBody(OpenAIEndpointResponseBody):
    id: str = Field(description="The unique identifier for the moderation request.")
    model: str = Field(description="The model used to generate the moderation results.")
    results: list[Result] = Field(description="A list of moderation objects.")



from typing import List, Literal, Optional

from pydantic import Field

from ..data_models import OpenAIEndpointQueryParam, OpenAIEndpointResponseBody
from .batch_models import OpenAIBatchResponseBody


class OpenAIListBatchQueryParam(OpenAIEndpointQueryParam):
    after: str | None = Field(None, description="A cursor for use in pagination. ")

    limit: int | None = Field(
        default=20,
        description="A limit on the number of objects to be returned. "
        "Limit can range between 1 and 100, and the default is 20.",
    )


class OpenAIListBatchResponseBody(OpenAIEndpointResponseBody):
    data: list[OpenAIBatchResponseBody] = Field(
        description="The list of batch objects."
    )

    object: Literal["list"] = Field(
        description='The object type, which is always "list".'
    )

    first_id: str = Field(description="The first object id in the list")

    last_id: str = Field(description="The last object id in the list")

    has_more: bool = Field(
        description="Whether there are more results " "available after this batch."
    )



from ..typing.pydantic_ import ConfigDict, Field, FieldInfo, field_validator
from ..typing.typing_ import UNDEFINED, Any, Callable
from .schema_model import SchemaModel, common_config


class FieldModel(SchemaModel):
    """Model for defining and managing field definitions.

    Provides a structured way to define fields with:
    - Type annotations and validation
    - Default values and factories
    - Documentation and metadata
    - Serialization options

    Example:
        ```python
        field = FieldModel(
            name="age",
            annotation=int,
            default=0,
            description="User age in years",
            validator=lambda v: v if v >= 0 else 0
        )
        ```

    Attributes:
        default: Default field value
        default_factory: Function to generate default value
        title: Field title for documentation
        description: Field description
        examples: Example values
        validators: Validation functions
        exclude: Exclude from serialization
        deprecated: Mark as deprecated
        frozen: Mark as immutable
        alias: Alternative field name
        alias_priority: Priority for alias resolution
        name: Field name (required)
        annotation: Type annotation
        validator: Validation function
        validator_kwargs: Validator parameters

    Notes:
        - All attributes except 'name' can be UNDEFINED
        - validator_kwargs are passed to field_validator decorator
        - Cannot have both default and default_factory
    """

    model_config = ConfigDict(extra="allow", validate_default=False, **common_config)

    # Field configuration attributes
    default: Any = UNDEFINED  # Default value
    default_factory: Callable = UNDEFINED  # Factory function for default value
    title: str = UNDEFINED  # Field title
    description: str = UNDEFINED  # Field description
    examples: list = UNDEFINED  # Example values
    validators: list = UNDEFINED  # Validation functions
    exclude: bool = UNDEFINED  # Exclude from serialization
    deprecated: bool = UNDEFINED  # Mark as deprecated
    frozen: bool = UNDEFINED  # Mark as immutable
    alias: str = UNDEFINED  # Alternative field name
    alias_priority: int = UNDEFINED  # Priority for alias resolution

    # Core field attributes
    name: str = Field(..., exclude=True)  # Field name (required)
    annotation: type | Any = Field(UNDEFINED, exclude=True)  # Type annotation
    validator: Callable | Any = Field(UNDEFINED, exclude=True)  # Validation function
    validator_kwargs: dict | Any = Field(
        default_factory=dict, exclude=True
    )  # Validator parameters

    @property
    def field_info(self) -> FieldInfo:
        """Generate Pydantic FieldInfo object from field configuration.

        Returns:
            FieldInfo object with all configured attributes

        Notes:
            - Uses clean dict to exclude UNDEFINED values
            - Sets annotation to Any if not specified
            - Preserves all metadata in field_info
        """
        annotation = self.annotation if self.annotation is not UNDEFINED else Any
        field_obj: FieldInfo = Field(**self.to_dict(True))  # type: ignore
        field_obj.annotation = annotation
        return field_obj

    @property
    def field_validator(self) -> dict[str, Callable] | None:
        """Generate field validator configuration.

        Returns:
            Dictionary mapping validator name to function,
            or None if no validator defined

        Notes:
            - Validator name is f"{field_name}_validator"
            - Uses validator_kwargs if provided
            - Returns None if validator is UNDEFINED
        """
        if self.validator is UNDEFINED:
            return None
        kwargs = self.validator_kwargs or {}
        return {
            f"{self.name}_validator": field_validator(self.name, **kwargs)(
                self.validator
            )
        }


__all__ = ["FieldModel"]



from collections.abc import Callable
from pathlib import Path
from typing import Any, Literal

from pydantic import BaseModel, Field


class ChunkContentParams(BaseModel):
    """Parameters for chunking text content into smaller pieces.

    This model defines parameters used by the chunk_content function to split text
    content into chunks, with support for both character-based and token-based chunking.
    """

    content: str = Field(description="The text content to be chunked")
    chunk_by: Literal["chars", "tokens"] = Field(
        default="chars",
        description="Method to use for chunking: 'chars' for character-based or 'tokens' for token-based",
    )
    tokenizer: Callable[[str], list[str]] | None = Field(
        default=None,
        description="Function to use for tokenization. Defaults to str.split if None",
    )
    chunk_size: int = Field(
        default=1024, ge=1, description="Target size for each chunk"
    )
    overlap: float = Field(
        default=0.0,
        ge=0.0,
        le=1.0,
        description="Fraction of overlap between chunks (0.0 to 1.0)",
    )
    threshold: int = Field(
        default=256, ge=0, description="Minimum size threshold for the last chunk"
    )
    metadata: dict[str, Any] = Field(
        default_factory=dict,
        description="Additional metadata to include with each chunk",
    )
    return_tokens: bool = Field(
        default=False,
        description="If True, return chunks as token lists; if False, return as joined strings",
    )


class CreatePathParams(BaseModel):
    """Parameters for creating file paths with various options.

    This model defines parameters used by the create_path function to generate
    file paths with support for timestamps, random hashes, and existence checks.
    """

    directory: Path | str = Field(
        description="Directory where the file will be created"
    )
    filename: str = Field(description="Base name of the file to create")
    extension: str | None = Field(
        default=None, description="File extension (with or without leading dot)"
    )
    timestamp: bool = Field(
        default=False, description="If True, adds a timestamp to the filename"
    )
    dir_exist_ok: bool = Field(
        default=True,
        description="If True, doesn't raise an error if the directory exists",
    )
    file_exist_ok: bool = Field(
        default=False, description="If True, allows overwriting of existing files"
    )
    time_prefix: bool = Field(
        default=False, description="If True, adds timestamp as prefix instead of suffix"
    )
    timestamp_format: str | None = Field(
        default=None, description="Custom format for the timestamp (strftime format)"
    )
    random_hash_digits: int = Field(
        default=0,
        ge=0,
        description="Number of random hash digits to append to filename",
    )


class DirToFilesParams(BaseModel):
    """Parameters for processing directories and collecting file paths.

    This model defines parameters used by the dir_to_files function to recursively
    process directories and collect matching file paths.
    """

    directory: Path | str = Field(description="Directory to process recursively")
    file_types: list[str] | None = Field(
        default=None,
        description="List of file extensions to include (e.g., ['.txt', '.pdf']). If None, includes all types",
    )
    max_workers: int | None = Field(
        default=None,
        description="Maximum number of worker threads for concurrent processing",
    )
    ignore_errors: bool = Field(
        default=False,
        description="If True, log warnings for errors instead of raising exceptions",
    )
    verbose: bool = Field(
        default=False, description="If True, print verbose output during processing"
    )


class FileToChunksParams(BaseModel):
    """Parameters for splitting file content into chunks.

    This model defines parameters used by the file_to_chunks function to read
    and split file content into chunks with metadata.
    """

    file_path: Path | str = Field(description="Path to the file to be processed")
    chunk_func: Callable[[str, int, float, int], list[str]] = Field(
        description="Function to use for chunking the content"
    )
    chunk_size: int = Field(
        default=1500, ge=1, description="Target size for each chunk"
    )
    overlap: float = Field(
        default=0.1,
        ge=0.0,
        le=1.0,
        description="Fraction of overlap between chunks (0.0 to 1.0)",
    )
    threshold: int = Field(
        default=200, ge=0, description="Minimum size threshold for the last chunk"
    )
    encoding: str = Field(
        default="utf-8", description="File encoding to use when reading the file"
    )
    custom_metadata: dict[str, Any] | None = Field(
        default=None, description="Additional metadata to include with each chunk"
    )
    output_dir: Path | str | None = Field(
        default=None, description="Directory to save output chunks (if provided)"
    )
    verbose: bool = Field(
        default=False, description="If True, print verbose output during processing"
    )
    timestamp: bool = Field(
        default=True, description="If True, include timestamp in output filenames"
    )
    random_hash_digits: int = Field(
        default=4,
        ge=0,
        description="Number of random hash digits to include in output filenames",
    )


class SaveToFileParams(BaseModel):
    """Parameters for saving text content to a file.

    This model defines parameters used by the save_to_file function to save
    text content to a file with various options for naming and handling.
    """

    text: str = Field(description="The text content to save to file")
    directory: Path | str = Field(description="Directory where the file will be saved")
    filename: str = Field(description="Name of the file to create")
    extension: str | None = Field(
        default=None, description="File extension (with or without leading dot)"
    )
    timestamp: bool = Field(
        default=False, description="If True, append timestamp to filename"
    )
    dir_exist_ok: bool = Field(
        default=True, description="If True, creates directory if it doesn't exist"
    )
    file_exist_ok: bool = Field(
        default=False, description="If True, allows overwriting existing files"
    )
    time_prefix: bool = Field(
        default=False, description="If True, prepend timestamp instead of append"
    )
    timestamp_format: str | None = Field(
        default=None, description="Custom format for timestamp (strftime format)"
    )
    random_hash_digits: int = Field(
        default=0,
        ge=0,
        description="Number of random hash digits to append to filename",
    )
    verbose: bool = Field(
        default=True, description="If True, logs the file path after saving"
    )



from typing import Any

from pydantic import Field

from ..data_models import OllamaEndpointRequestBody, OllamaEndpointResponseBody


class OllamaPushModelRequestBody(OllamaEndpointRequestBody):
    name: str = Field(
        description="Name of the model to push in the form of '<namespace>/<model>:<tag>'"
    )

    insecure: Any = Field(
        None,
        description="Allow insecure connections to the library. "
        "Only use this if you are pushing to your library during development.",
    )

    stream: bool = Field(
        True,
        description="If 'false' the response will be returned as a single response object, "
        "rather than a stream of objects",
    )


class OllamaPushModelResponseBody(OllamaEndpointResponseBody):
    status: str = Field(None)

    digest: str = Field(None)

    total: int = Field(None)



"""
Copyright 2024 HaiyangLi

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from enum import Enum
from typing import Any

from pydantic import BaseModel, Field

from lion import Branch
from lion.protocols.operatives.instruct import Instruct

from .prompt import PROMPT
from .utils import parse_selection, parse_to_representation


class SelectionModel(BaseModel):
    """Model representing the selection output."""

    selected: list[Any] = Field(default_factory=list)


async def select(
    instruct: Instruct | dict[str, Any],
    choices: list[str] | type[Enum] | dict[str, Any],
    max_num_selections: int = 1,
    branch: Branch | None = None,
    branch_kwargs: dict[str, Any] | None = None,
    return_branch: bool = False,
    verbose: bool = False,
    **kwargs: Any,
) -> SelectionModel | tuple[SelectionModel, Branch]:
    """Perform a selection operation from given choices.

    Args:
        instruct: Instruction model or dictionary.
        choices: Options to select from.
        max_num_selections: Maximum selections allowed.
        branch: Existing branch or None to create a new one.
        branch_kwargs: Additional arguments for branch creation.
        return_branch: If True, return the branch with the selection.
        verbose: Whether to enable verbose output.
        **kwargs: Additional keyword arguments.

    Returns:
        A SelectionModel instance, optionally with the branch.
    """
    if verbose:
        print(f"Starting selection with up to {max_num_selections} choices.")

    branch = branch or Branch(**(branch_kwargs or {}))
    selections, contents = parse_to_representation(choices)
    prompt = PROMPT.format(max_num_selections=max_num_selections, choices=selections)

    if isinstance(instruct, Instruct):
        instruct = instruct.to_dict()

    instruct = instruct or {}

    if instruct.get("instruction", None) is not None:
        instruct["instruction"] = f"{instruct['instruction']}\n\n{prompt} \n\n "
    else:
        instruct["instruction"] = prompt

    context = instruct.get("context", None) or []
    context = [context] if not isinstance(context, list) else context
    context.extend([{k: v} for k, v in zip(selections, contents)])
    instruct["context"] = context

    response_model: SelectionModel = await branch.operate(
        operative_model=SelectionModel,
        **kwargs,
        **instruct,
    )
    if verbose:
        print(f"Received selection: {response_model.selected}")

    selected = response_model
    if isinstance(response_model, BaseModel) and hasattr(response_model, "selected"):
        selected = response_model.selected
    selected = [selected] if not isinstance(selected, list) else selected

    corrected_selections = [parse_selection(i, choices) for i in selected]

    if isinstance(response_model, BaseModel):
        response_model.selected = corrected_selections

    elif isinstance(response_model, dict):
        response_model["selected"] = corrected_selections

    if return_branch:
        return response_model, branch
    return response_model



from .cancel_batch import OpenAICancelBatchPathParam
from .create_batch import OpenAIBatchRequestBody
from .list_batch import OpenAIListBatchQueryParam
from .request_object_models import (
    OpenAIBatchRequestInputObject,
    OpenAIBatchRequestOutputObject,
)
from .retrieve_batch import OpenAIRetrieveBatchPathParam

__all__ = [
    "OpenAIBatchRequestBody",
    "OpenAIRetrieveBatchPathParam",
    "OpenAICancelBatchPathParam",
    "OpenAIListBatchQueryParam",
    "OpenAIBatchRequestInputObject",
    "OpenAIBatchRequestOutputObject",
]



from typing import List, Literal, Optional

from pydantic import Field

from ..data_models import OpenAIEndpointQueryParam, OpenAIEndpointResponseBody
from .fine_tuning_job_models import OpenAIFineTuningJobResponseBody


class OpenAIListFineTuningJobsQueryParam(OpenAIEndpointQueryParam):
    after: str | None = Field(
        None,
        description="Identifier for the last job from the previous pagination request.",
    )

    limit: int | None = Field(
        20, description="Number of fine-tuning jobs to retrieve.", ge=1
    )


class OpenAIListFineTuningJobsResponseBody(OpenAIEndpointResponseBody):
    object: Literal["list"] = Field(
        description="The object type, always 'list' for a list of fine-tuning jobs.",
    )

    data: list[OpenAIFineTuningJobResponseBody] = Field(
        description="A list of fine-tuning job events."
    )

    has_more: bool = Field(
        description="Whether there are more results available after this batch."
    )



from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Field


class Role(str, Enum):
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"
    TOOL = "tool"


class Message(BaseModel):
    role: Role = Field(
        description="The role of the message, either 'system', 'user', 'assistant', or 'tool'"
    )

    content: str = Field(description="The content of the message")

    images: list[str] | None = Field(
        None,
        description="a list of images to include in the message (for multimodal models such as 'llava')",
    )

    tool_calls: list | None = Field(
        None, description="a list of tools the model wants to use"
    )



from pydantic import Field

from ..data_models import OpenAIEndpointPathParam, OpenAIEndpointRequestBody


class OpenAIUploadPartPathParam(OpenAIEndpointPathParam):
    upload_id: str = Field(description="The ID of the Upload.")


class OpenAIUploadPartRequestBody(OpenAIEndpointRequestBody):
    data: bytes = Field(description="The chunk of bytes for this Part.")



from typing import Literal

from pydantic import BaseModel, Field


class Function(BaseModel):
    name: str = Field(description="The name of the function to call.")


class ToolChoice(BaseModel):
    type: Literal["function"] = Field(
        description="The type of the tool. Currently," " only function is supported."
    )
    function: Function = Field(description="Specifies the function to be called.")



from lion.core.typing import (
    ID,
    Any,
    BaseModel,
    Condition,
    ConfigDict,
    Field,
    Note,
    Relational,
    field_serializer,
)

from .element import Element


class EdgeCondition(BaseModel, Condition):
    """Represents a condition associated with an edge in the Lion framework.

    This class combines Condition characteristics with Pydantic's
    BaseModel for robust data validation and serialization.

    Attributes:
        source (Any): The source for condition evaluation.
    """

    source: Any = Field(
        default=None,
        title="Source",
        description="The source for condition evaluation",
    )

    model_config = ConfigDict(
        extra="allow",
        arbitrary_types_allowed=True,
    )


class Edge(Element):
    """
    Represents an edge in a graph structure.

    An edge connects a head node to a tail node and can store additional
    properties such as conditions and labels. It is used to model relationship
    between nodes in graph-based data structures.

    Attributes:
        properties (Note): Stores additional properties of the edge. This can
            include conditions that must be met for the edge to be traversed
            and labels that describe the edge.
        head (str): The ID of the head node in the graph.
        tail (str): The ID of the tail node in the graph.
    """

    properties: Note = Field(default_factory=Note)
    head: ID[Relational].ID
    tail: ID[Relational].ID

    def __init__(
        self,
        head: ID[Relational].Ref,
        tail: ID[Relational].Ref,
        condition: EdgeCondition | None = None,
        label: list[str] | None = None,
        **kwargs,
    ):
        """
        Initialize an Edge.

        This constructor sets up an edge by linking a head node to a tail node,
        with optional conditions and labels. Additional properties can also be
        provided via keyword arguments.

        Args:
            head (Relational | str): The head node or its ID. This is the
                starting point of the edge.
            tail (Relational | str): The tail node or its ID. This is the end
                point of the edge.
            condition (EdgeCondition | None): An optional condition that must
                be satisfied for the edge to be traversed.
            label (list[str] | None): An optional list of labels that describe
                the edge.
            kwargs: Optional additional properties for the edge.
        """
        head = ID.get_id(head)
        tail = ID.get_id(tail)

        super().__init__(head=head, tail=tail)

        # Initialize properties
        if condition:
            self.properties.content["condition"] = condition
        if label:
            self.properties.content["label"] = label
        for key, value in kwargs.items():
            self.properties.content[key] = value

    @field_serializer("properties")
    def _serialize_properties(self, value: Note) -> dict[str, Any]:
        """
        Serialize the properties of the edge.

        This method converts the `properties` field into a format suitable for
        storage or transmission, typically as part of serialization to JSON or
        other formats.

        Args:
            value: The properties Note to serialize.

        Returns:
            The serialized properties.
        """
        return value.content

    async def check_condition(self, *args, **kwargs) -> bool:
        """
        Check if the edge's condition is satisfied.

        This method checks whether the edge can be traversed by evaluating its
        condition. If no condition is specified, the edge is considered
        traversable by default.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            bool: True if the condition is satisfied or if no condition exists,
                  False otherwise.
        """
        condition: EdgeCondition | None = self.properties.get("condition", None)
        if condition:
            return await condition.apply(*args, **kwargs)
        return True  # If no condition exists, the edge is always traversable

    def update_property(self, key: str, value: Any) -> None:
        """
        Update a property value directly.

        Args:
            key: The property key to update.
            value: The new value for the property.
        """
        self.properties.content[key] = value

    def update_condition_source(self, source: dict) -> None:
        """
        Update the condition source.

        Args:
            source: The new source dictionary for the condition.
        """
        condition = self.properties.get("condition")
        if condition:
            condition.source = source


__all__ = ["Edge"]



from typing import List, Literal, Optional

from pydantic import Field

from ..data_models import (
    OpenAIEndpointPathParam,
    OpenAIEndpointQueryParam,
    OpenAIEndpointResponseBody,
)
from .fine_tuning_job_event_models import OpenAIFineTuningJobEventResponseBody


# Request model
class OpenAIListFineTuningEventsPathParam(OpenAIEndpointPathParam):
    fine_tuning_job_id: str = Field(
        description="The ID of the fine-tuning job to get events for."
    )


class OpenAIListFineTuningEventsQueryParam(OpenAIEndpointQueryParam):
    after: str | None = Field(
        None,
        description="Identifier for the last event from the previous pagination request.",
    )

    limit: int | None = Field(20, description="Number of events to retrieve.", ge=1)


class OpenAIListFineTuningEventsResponseBody(OpenAIEndpointResponseBody):
    object: Literal["list"] = Field(
        description="The object type, always 'list' for a list of fine-tuning events.",
    )

    data: list[OpenAIFineTuningJobEventResponseBody] = Field(
        description="A list of fine-tuning events."
    )

    has_more: bool = Field(
        description="Whether there are more results available after this batch."
    )



from lion.core.typing import ID, UNDEFINED, Any, BaseModel, IDError, Literal, LnID, Note
from lion.integrations.pydantic_ import break_down_pydantic_annotation
from lion.libs.utils import time

DEFAULT_SYSTEM = "You are a helpful AI assistant. Let's think step by step."


def format_system_content(
    system_datetime: bool | str | None,
    system_message: str,
) -> Note:
    """
    Format system message content with optional datetime information.

    Args:
        system_datetime: Flag or string for datetime inclusion
        system_message: The system message content

    Returns:
        Note: Formatted system content
    """
    content = Note(system=str(system_message or DEFAULT_SYSTEM))
    if system_datetime:
        if isinstance(system_datetime, str):
            content["system_datetime"] = system_datetime
        else:
            content["system_datetime"] = time(type_="iso", timespec="minutes")
    return content


def prepare_request_response_format(request_fields: dict) -> str:
    """
    Prepare a standardized format for request responses.

    Args:
        request_fields: Dictionary of fields to include in response

    Returns:
        str: Formatted response template
    """
    return (
        "**MUST RETURN JSON-PARSEABLE RESPONSE ENCLOSED BY JSON CODE BLO"
        f"CKS.** \n```json\n{request_fields}\n```"
    ).strip()


def format_image_item(idx: str, x: str, /) -> dict[str, Any]:
    """
    Create an image_url dictionary for content formatting.

    Args:
        idx: Base64 encoded image data
        x: Image detail level

    Returns:
        dict: Formatted image item
    """
    return {
        "type": "image_url",
        "image_url": {
            "url": f"data:image/jpeg;base64,{idx}",
            "detail": x,
        },
    }


def format_text_item(item: Any) -> str:
    """
    Format a text item or list of items into a string.

    Args:
        item: Text item(s) to format

    Returns:
        str: Formatted text
    """
    msg = ""
    item = [item] if not isinstance(item, list) else item
    for j in item:
        if isinstance(j, dict):
            for k, v in j.items():
                msg += f"- {k}: {v} \n\n"
        else:
            msg += f"{j}\n"
    return msg


def format_text_content(content: dict) -> str:
    """
    Format dictionary content into a structured text format.

    Args:
        content: Dictionary containing content sections

    Returns:
        str: Formatted text content
    """
    if "plain_content" in content and isinstance(content["plain_content"], str):
        return content["plain_content"]

    msg = "\n---\n # Task\n"
    for k, v in content.items():
        if k in [
            "guidance",
            "instruction",
            "context",
            "request_response_format",
            "tool_schemas",
        ]:
            if k == "request_response_format":
                k = "response format"
            msg += f"## **Task {k}**\n{format_text_item(v)}\n\n"
    msg += "\n\n---\n"
    return msg


def format_image_content(
    text_content: str,
    images: list,
    image_detail: Literal["low", "high", "auto"],
) -> dict[str, Any]:
    """
    Format text content with images for message content.

    Args:
        text_content: The text content to format
        images: List of images to include
        image_detail: Level of detail for images

    Returns:
        dict: Formatted content with text and images
    """
    content = [{"type": "text", "text": text_content}]
    content.extend(format_image_item(i, image_detail) for i in images)
    return content


def prepare_instruction_content(
    guidance: str | None = None,
    instruction: str | None = None,
    context: str | dict | list | None = None,
    request_fields: dict | list[str] | None = None,
    plain_content: str | None = None,
    request_model: BaseModel = None,
    images: str | list | None = None,
    image_detail: Literal["low", "high", "auto"] | None = None,
    tool_schemas: dict | None = None,
) -> Note:
    """
    Prepare the content for an instruction message.

    Args:
        guidance: Optional guidance text
        instruction: Main instruction content
        context: Additional context information
        request_fields: Fields to request in response
        plain_content: Plain text content
        request_model: Pydantic model for structured requests
        images: Images to include
        image_detail: Level of detail for images
        tool_schemas: Tool schemas to include

    Returns:
        Note: Prepared instruction content

    Raises:
        ValueError: If both request_fields and request_model are provided
    """
    if request_fields and request_model:
        raise ValueError("only one of request_fields or request_model can be provided")

    out_ = {"context": []}
    if guidance:
        out_["guidance"] = guidance
    if instruction:
        out_["instruction"] = instruction
    if context:
        if isinstance(context, list):
            out_["context"].extend(context)
        else:
            out_["context"].append(context)
    if images:
        out_["images"] = images if isinstance(images, list) else [images]
        out_["image_detail"] = image_detail or "low"

    if tool_schemas:
        out_["tool_schemas"] = tool_schemas

    if request_model:
        out_["request_model"] = request_model
        request_fields = break_down_pydantic_annotation(request_model)
        out_["context"].append(
            {"respond_schema_info": request_model.model_json_schema()}
        )

    if request_fields:
        _fields = request_fields if isinstance(request_fields, dict) else {}
        if not isinstance(request_fields, dict):
            _fields = {i: "..." for i in request_fields}
        out_["request_fields"] = _fields
        out_["request_response_format"] = prepare_request_response_format(
            request_fields=_fields
        )

    if plain_content:
        out_["plain_content"] = plain_content

    return Note(
        **{k: v for k, v in out_.items() if v not in [None, UNDEFINED]},
    )


def validate_sender_recipient(
    value: Any, /
) -> LnID | Literal["system", "user", "N/A", "assistant"]:
    """
    Validate sender and recipient fields for mail-like communication.

    Args:
        value: The value to validate

    Returns:
        Union[LnID, Literal]: Valid sender/recipient value

    Raises:
        ValueError: If value is not a valid sender or recipient
    """
    if value in ["system", "user", "N/A", "assistant"]:
        return value

    if value is None:
        return "N/A"

    try:
        return ID.get_id(value)
    except IDError as e:
        raise ValueError("Invalid sender or recipient") from e



from collections import deque
from datetime import UTC, datetime, timezone

from pydantic import BaseModel, Field

from .complete_request_info import CompleteRequestInfo, CompleteRequestTokenInfo


class RateLimiter(BaseModel):
    limit_tokens: int = Field(default=None)

    limit_requests: int = Field(default=None)

    remaining_tokens: int = Field(default=None)

    remaining_requests: int = Field(default=None)

    last_check_timestamp: float = Field(
        default=None, description="Last time to check tokens and requests."
    )

    unreleased_requests: deque = Field(
        default_factory=deque,
        description="completed request info for replenish",
        exclude=True,
    )

    def append_complete_request_token_info(self, info: CompleteRequestInfo):
        if not self.limit_tokens and not self.limit_requests:
            # no limits
            return

        self.unreleased_requests.append(info)
        if self.limit_tokens and isinstance(info, CompleteRequestTokenInfo):
            # For request with token limits only
            if self.remaining_tokens:
                self.remaining_tokens -= info.token_usage
            else:
                self.remaining_tokens = self.limit_tokens - info.token_usage

        if self.limit_requests:
            if self.remaining_requests:
                self.remaining_requests -= 1
            else:
                self.remaining_requests = self.limit_requests - 1

    def release_tokens(self):
        self.last_check_timestamp = datetime.now(UTC).timestamp()
        while self.unreleased_requests:
            if (
                datetime.now(UTC).timestamp() - self.unreleased_requests[0].timestamp
                > 60
            ):
                release_info = self.unreleased_requests.popleft()
                if (
                    isinstance(release_info, CompleteRequestTokenInfo)
                    and self.remaining_tokens
                ):
                    self.remaining_tokens += release_info.token_usage
                if self.remaining_requests:
                    self.remaining_requests += 1
            else:
                break

    def update_rate_limit(self, request_datetime_header, total_token_usage: int = None):
        # rate limiter tokens only tracks if there are token usage info
        # otherwise, tracks requests num
        date_format = "%a, %d %b %Y %H:%M:%S GMT"  # the format of the date string according to RFC 1123
        # (in http response header)
        dt = datetime.strptime(request_datetime_header, date_format)
        dt = dt.replace(tzinfo=timezone.utc)
        request_timestamp = dt.timestamp()

        if total_token_usage:
            complete_request_info = CompleteRequestTokenInfo(
                timestamp=request_timestamp, token_usage=total_token_usage
            )
        else:
            complete_request_info = CompleteRequestInfo(timestamp=request_timestamp)
        self.append_complete_request_token_info(complete_request_info)

    def check_availability(
        self, request_token_len: int = 0, estimated_output_len: int = 0
    ):
        if self.remaining_tokens is not None:
            if request_token_len + estimated_output_len > self.remaining_tokens:
                return False
        if self.remaining_requests is not None:
            if self.remaining_requests <= 0:
                return False
        return True


class RateLimitError(Exception):
    def __init__(self, message, input_token_len, estimated_output_len):
        super().__init__(message)
        self.requested_tokens = input_token_len + estimated_output_len



from lion.core.generic.component import Component
from lion.core.generic.element import Element
from lion.core.generic.log import Log
from lion.core.generic.log_manager import LogManager
from lion.core.generic.pile import Pile
from lion.core.generic.progression import Progression
from lion.core.generic.utils import to_list_type

__all__ = [
    "Component",
    "Flow",
    "Log",
    "Pile",
    "Progression",
    "Node",
    "to_list_type",
    "Element",
    "LogManager",
]



from typing import Literal

from pydantic import BaseModel, Field

from ..data_models import OpenAIEndpointResponseBody


class Metrics(BaseModel):
    step: int = Field(description="The step number associated with these metrics.")

    train_loss: float = Field(description="The training loss at this step.")

    train_mean_token_accuracy: float = Field(
        description="The mean token accuracy during training at this step."
    )
    valid_loss: float = Field(description="The validation loss at this step.")

    valid_mean_token_accuracy: float = Field(
        description="The mean token accuracy during validation at this step."
    )

    full_valid_loss: float = Field(description="The full validation loss at this step.")

    full_valid_mean_token_accuracy: float = Field(
        description="The full mean token accuracy during validation at this step."
    )


class OpenAIFineTuningJobCheckpointResponseBody(OpenAIEndpointResponseBody):
    id: str = Field(description="The ID of the checkpoint.")

    created_at: int = Field(
        description="The Unix timestamp (in seconds) of when the checkpoint was created.",
    )

    fine_tuned_model_checkpoint: str = Field(
        description="The name of the fine-tuned model checkpoint."
    )

    step_number: int = Field(
        description="The step number of this checkpoint in the fine-tuning process.",
    )

    metrics: Metrics = Field(description="Metrics associated with the checkpoint.")

    fine_tuning_job_id: str = Field(
        description="The ID of the fine-tuning job this checkpoint belongs to."
    )

    object: Literal["fine_tuning.job.checkpoint"] = Field(
        description="The object type, always 'fine_tuning.job.checkpoint'.",
    )



from typing import Literal

from pydantic import BaseModel, Field

from ..data_models import OpenAIEndpointResponseBody


class Data(BaseModel):
    code: str = Field(description="An error code identifying the error type.")

    message: str = Field(
        description="A human-readable message providing more details"
        " about the error."
    )

    param: str | None = Field(
        None,
        description="The name of the parameter that caused the "
        "error, if applicable.",
    )

    line: int | None = Field(
        None,
        description="The line number of the input file where "
        "the error occurred, if applicable.",
    )


class Error(BaseModel):
    object: Literal["list"] = Field(
        description="The object type, which is always 'list'."
    )

    data: Data = Field(description="A list of error data.")


class RequestCounts(BaseModel):
    total: int = Field(description="Total number of requests in the batch.")

    completed: int = Field(
        description="Number of requests that have been completed successfully."
    )

    failed: int = Field(description="Number of requests that have failed.")


class OpenAIBatchResponseBody(OpenAIEndpointResponseBody):
    id: str = Field(description="A unique identifier for the batch.")

    object: Literal["batch"] = Field(
        description="The object type, which is always 'batch'."
    )

    endpoint: str = Field(description="The API endpoint used for this batch.")

    errors: Error | None = Field(
        None, description="Errors encountered during batch processing."
    )

    input_file_id: str | None = Field(
        None, description="The ID of the input file for the batch."
    )

    completion_window: str | None = Field(
        None,
        description="The time frame within which the batch " "should be processed.",
    )

    status: str = Field(description="The current status of the batch.")

    output_file_id: str | None = Field(
        None,
        description="The ID of the file containing the outputs of "
        "successfully executed requests.",
    )

    error_file_id: str | None = Field(
        None,
        description="The ID of the file containing the outputs "
        "of requests with errors.",
    )

    created_at: int | None = Field(
        None,
        description="The Unix timestamp (in seconds) for when "
        "the batch was created.",
    )

    in_progress_at: int | None = Field(
        None,
        description="The Unix timestamp (in seconds) for when the"
        " batch started processing.",
    )

    expires_at: int | None = Field(
        None,
        description="The Unix timestamp (in seconds) for"
        " when the batch will expire.",
    )

    finalizing_at: int | None = Field(
        None,
        description="The Unix timestamp (in seconds) for"
        " when the batch started finalizing.",
    )

    completed_at: int | None = Field(
        None,
        description="The Unix timestamp (in seconds) for"
        " when the batch was completed.",
    )

    failed_at: int | None = Field(
        None,
        description="The Unix timestamp (in seconds) for" " when the batch failed.",
    )

    expired_at: int | None = Field(
        None,
        description="The Unix timestamp (in seconds) for" " when the batch expired.",
    )

    cancelling_at: int | None = Field(
        None,
        description="The Unix timestamp (in seconds) for"
        " when the batch started cancelling.",
    )

    cancelled_at: int | None = Field(
        None,
        description="The Unix timestamp (in seconds) for"
        " when the batch was cancelled.",
    )

    request_counts: RequestCounts = Field(
        description="The request counts for different statuses" " within the batch."
    )

    metadata: dict | None = Field(
        description="Set of 16 key-value pairs that can be" " attached to an object."
    )



from lion.libs.utils import insert_random_hyphens, unique_hash
from lion.settings import LionIDConfig, Settings

from .concepts import Container, Generic, ItemError, Observable, Ordering, T
from .pydantic_ import Field
from .typing_ import Annotated, Container, Mapping, Sequence, TypeAlias

LnID: TypeAlias = Annotated[str, Field(description="A unique identifier.")]


class IDError(ItemError):
    """Exception raised when an item does not have a Lion ID."""

    def __init__(
        self,
        message: str = "Item must contain a Lion ID.",
        item_id: str | None = None,
    ):
        super().__init__(message, item_id)


class ID(Generic[T]):
    """
    A generic class that provides ID-related functionality for Observable objects.

    This class handles the generation, validation, and management of unique identifiers
    within the Lion system. It provides type aliases for various ID-related operations
    and methods for working with IDs.
    """

    # For functions that accept either ID or item
    Ref: TypeAlias = LnID | T  # type: ignore

    # For functions requiring just the ID
    ID: TypeAlias = LnID

    # For functions requiring Observable object
    Item = T  # type: ignore

    # For collections
    IDSeq: TypeAlias = Sequence[LnID] | Ordering[LnID]
    ItemSeq: TypeAlias = (  # type: ignore
        Sequence[T] | Mapping[LnID, T] | Container[LnID | T]
    )
    RefSeq: TypeAlias = IDSeq | ItemSeq

    # For system-level interactions
    SenderRecipient: TypeAlias = LnID | T  # type: ignore

    @staticmethod
    def id(
        config: LionIDConfig = Settings.Config.ID,
        n: int = None,
        prefix: str = None,
        postfix: str = None,
        random_hyphen: bool = None,
        num_hyphens: int = None,
        hyphen_start_index: int = None,
        hyphen_end_index: int = None,
    ) -> LnID:
        """
        Generate a unique identifier.

        Args:
            n: Length of the ID (excluding prefix and postfix).
            prefix: String to prepend to the ID.
            postfix: String to append to the ID.
            random_hyphen: If True, insert random hyphens into the ID.
            num_hyphens: Number of hyphens to insert if random_hyphen is True.
            hyphen_start_index: Start index for hyphen insertion.
            hyphen_end_index: End index for hyphen insertion.

        Returns:
            A unique identifier string.
        """
        _dict = {
            "n": n,
            "prefix": prefix,
            "postfix": postfix,
            "random_hyphen": random_hyphen,
            "num_hyphens": num_hyphens,
            "hyphen_start_index": hyphen_start_index,
            "hyphen_end_index": hyphen_end_index,
        }
        _dict = {k: v for k, v in _dict.items() if v is not None}
        config = {**config.to_dict(), **_dict}
        return ID._id(**config)

    @staticmethod
    def _id(
        *,
        n: int,
        prefix: str = "",
        postfix: str = "",
        random_hyphen: bool = False,
        num_hyphens: int = 0,
        hyphen_start_index: int = 6,
        hyphen_end_index: int = -6,
    ):
        _id = unique_hash(n)
        if random_hyphen:
            _id = insert_random_hyphens(
                s=_id,
                num_hyphens=num_hyphens,
                start_index=hyphen_start_index,
                end_index=hyphen_end_index,
            )
        if prefix:
            _id = f"{prefix}{_id}"
        if postfix:
            _id = f"{_id}{postfix}"

        return _id

    @staticmethod
    def get_id(
        item,
        config: LionIDConfig = Settings.Config.ID,
        /,
    ) -> str:
        """
        Get the Lion ID of an item.

        Args:
            item: The item to get the ID from.
            config: Configuration dictionary for ID validation.

        Returns:
            The Lion ID of the item.

        Raises:
            LionIDError: If the item does not contain a valid Lion ID.
        """

        item_id = None
        if isinstance(item, Sequence) and len(item) == 1:
            item = item[0]

        if isinstance(item, Observable):
            item_id: str = item.ln_id
        else:
            item_id = item

        check = isinstance(item_id, str)
        if check:
            id_len = (
                (len(config.prefix) if config.prefix else 0)
                + config.n
                + config.num_hyphens
                + (len(config.postfix) if config.postfix else 0)
            )
            if len(item_id) != id_len:
                check = False
        if check and config.prefix:
            if item_id.startswith(config.prefix):
                item_id = item_id[len(config.prefix) :]  # noqa
            else:
                check = False
        if check and config.postfix:
            if item_id.endswith(config.postfix):
                item_id = item_id[: -len(config.postfix)]
            else:
                check = False
        if check and config.num_hyphens:
            if config.num_hyphens != item_id.count("-"):
                check = False
        if check and config.hyphen_start_index:
            idx = config.hyphen_start_index - len(config.prefix)
            if idx > 0 and "-" in item_id[:idx]:
                check = False
        if check and config.hyphen_end_index:
            if config.hyphen_end_index < 0:
                idx = config.hyphen_end_index + id_len
            idx -= len(config.prefix + config.postfix)
            if idx < 0 and "-" in item_id[idx:]:
                check = False

        if check:
            return config.prefix + item_id + config.postfix
        if (
            isinstance(item_id, str) and len(item_id) == 32
        ):  # for backward compatibility
            return item_id
        raise IDError(
            f"The input object of type <{type(item).__name__}> does "
            "not contain or is not a valid Lion ID. Item must be an instance"
            " of `Observable` or a valid `ln_id`."
        )

    @staticmethod
    def is_id(
        item,
        config: LionIDConfig = Settings.Config.ID,
        /,
    ) -> bool:
        """
        Check if an item is a valid Lion ID.

        Args:
            item: The item to check.
            config: Configuration dictionary for ID validation.

        Returns:
            True if the item is a valid Lion ID, False otherwise.
        """
        try:
            ID.get_id(item, config)
            return True
        except IDError:
            return False


__all__ = [
    "LnID",
    "ID",
    "IDError",
]



from .branch_config import BranchConfig, MessageConfig
from .id_config import LionIDConfig
from .imodel_config import iModelConfig
from .log_config import LogConfig
from .retry_config import RetryConfig, TimedFuncCallConfig

__all__ = [
    "LogConfig",
    "LionIDConfig",
    "RetryConfig",
    "TimedFuncCallConfig",
    "iModelConfig",
    "BranchConfig",
    "MessageConfig",
]



from pydantic import BaseModel, Field


class CompleteRequestInfo(BaseModel):
    timestamp: float = Field(description="HTTP response generated time")


class CompleteRequestTokenInfo(CompleteRequestInfo):
    token_usage: int = Field(description="Number of tokens used in the request")



from enum import Enum


class TTSModel(str, Enum):
    TTS_1 = "tts-1"
    TTS_1_HD = "tts-1-hd"


class Voice(str, Enum):
    ALLOY = "alloy"
    ECHO = "echo"
    FABLE = "fable"
    ONYX = "onyx"
    NOVA = "nova"
    SHIMMER = "shimmer"


class AudioFormat(str, Enum):
    MP3 = "mp3"
    OPUS = "opus"
    AAC = "aac"
    FLAC = "flac"
    WAV = "wav"
    PCM = "pcm"


class WhisperModel(str, Enum):
    WHISPER_1 = "whisper-1"


class TranscriptionResponseFormat(str, Enum):
    JSON = "json"
    TEXT = "text"
    SRT = "srt"
    VERBOSE_JSON = "verbose_json"
    VTT = "vtt"


class TimestampGranularity(str, Enum):
    WORD = "word"
    SEGMENT = "segment"



from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar


class Observable(ABC):
    """
    Abstract base class for objects that can be uniquely identified and tracked.

    All Observable objects must have a unique identifier (ln_id) that allows them
    to be tracked and referenced within the Lion system.
    """

    ln_id: str


T = TypeVar("T", bound=Observable)


class Container(ABC, Generic[T]):
    """
    Abstract base class for objects that can contain Observable items.

    Generic type T must be a subclass of Observable.
    """

    pass


class Ordering(ABC, Generic[T]):
    """
    Abstract base class for objects that maintain an ordered collection of items.

    Generic type T must be a subclass of Observable.
    """

    pass


class Communicatable(ABC):
    """
    Abstract base class for objects that can participate in communication.

    Defines the interface for objects that can send and receive messages within
    the Lion system.
    """

    pass


class Relational(Observable):

    pass


class Event(ABC):

    @abstractmethod
    async def invoke(self, *args, **kwargs) -> Any:
        pass


class Condition(ABC):

    @abstractmethod
    async def apply(self, *args, **kwargs) -> bool:
        pass


class Structure(ABC):
    pass


class ItemError(Exception):
    """Base exception for errors related to framework items."""

    def __init__(
        self,
        message: str = "Item error.",
        item_id: str | None = None,
    ):
        self.item_id = item_id
        item_info = f" Item ID: {item_id}" if item_id else ""
        super().__init__(f"{message}{item_info}")


class ItemNotFoundError(ItemError):
    """Exception raised when an item is not found."""

    def __init__(
        self,
        message: str = "Item not found.",
        item_id: str | None = None,
    ):
        super().__init__(message, item_id)


class ItemExistsError(ItemError):
    """Exception raised when an item already exists."""

    def __init__(
        self,
        message: str = "Item already exists.",
        item_id: str | None = None,
    ):
        super().__init__(message, item_id)


class RelationError(Exception):
    """Base exception for errors related to framework relations."""

    def __init__(self, message: str = "Relation error."):
        super().__init__(message)


__all__ = [
    "Observable",
    "Container",
    "Ordering",
    "Communicatable",
    "ItemError",
    "ItemNotFoundError",
    "ItemExistsError",
    "Relational",
    "Event",
    "Condition",
    "Structure",
    "RelationError",
]



from enum import Enum


class Detail(str, Enum):
    AUTO = "auto"
    LOW = "low"
    HIGH = "high"


class ServiceTier(str, Enum):
    AUTO = "auto"
    DEFAULT = "default"


class ToolChoice(str, Enum):
    NONE = "none"
    AUTO = "auto"
    REQUIRED = "required"



import numpy as np
from pydantic import BaseModel

from lion.core.generic import Log
from lion.core.typing import Note


class PerplexityTokenScore(BaseModel):
    token: str
    logprob: float


class PerplexityScores(BaseModel):

    completion_response: BaseModel
    original_tokens: list[str]
    n_samples: int

    @property
    def logprobs(self) -> list[float]:
        return [i.logprob for i in self.perplexity_scores]

    @property
    def predicted_tokens(self) -> list[str]:
        return [i.token for i in self.perplexity_scores]

    @property
    def perplexity(self) -> float:
        return np.exp(np.mean(self.logprobs))

    @property
    def perplexity_scores(self) -> list[PerplexityTokenScore]:
        outs = []
        log_prob = self.completion_response.choices[0].logprobs
        n = Note.from_dict(log_prob)
        for j in range(len(n[["content"]])):
            p_ = ["content", j]
            p_token = p_ + ["token"]
            p_logprob = p_ + ["logprob"]
            p_score = PerplexityTokenScore(token=n[p_token], logprob=n[p_logprob])
            outs.append(p_score)
        return outs

    @property
    def prompt_tokens(self) -> int:
        return self.completion_response.usage.prompt_tokens

    @property
    def completion_tokens(self) -> int:
        return self.completion_response.usage.completion_tokens

    def to_log(self) -> Log:
        dict_ = self.model_dump()
        info = dict_.pop("completion_response")
        dict_["perplexity"] = self.perplexity
        dict_["prompt_tokens"] = self.prompt_tokens
        dict_["completion_tokens"] = self.completion_tokens
        dict_["perplexity_scores"] = self.perplexity_scores

        return Log(content=dict_, info=info)



"""
Copyright 2024 HaiyangLi

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import asyncio
import functools
import time
from collections.abc import Callable
from typing import Any, TypeVar

from ..utils import time as _t
from .utils import force_async, is_coroutine_func

T = TypeVar("T")


class Throttle:
    """
    Provide a throttling mechanism for function calls.

    When used as a decorator, it ensures that the decorated function can only
    be called once per specified period. Subsequent calls within this period
    are delayed to enforce this constraint.

    Attributes:
        period: The minimum time period (in seconds) between successive calls.
    """

    def __init__(self, period: float) -> None:
        """
        Initialize a new instance of Throttle.

        Args:
            period: The minimum time period (in seconds) between
                successive calls.
        """
        self.period = period
        self.last_called = 0

    def __call__(self, func: Callable[..., T]) -> Callable[..., T]:
        """
        Decorate a synchronous function with the throttling mechanism.

        Args:
            func: The synchronous function to be throttled.

        Returns:
            The throttled synchronous function.
        """

        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            elapsed = _t() - self.last_called
            if elapsed < self.period:
                time.sleep(self.period - elapsed)
            self.last_called = _t()
            return func(*args, **kwargs)

        return wrapper

    def __call_async__(
        self, func: Callable[..., Callable[..., T]]
    ) -> Callable[..., Callable[..., T]]:
        """
        Decorate an asynchronous function with the throttling mechanism.

        Args:
            func: The asynchronous function to be throttled.

        Returns:
            The throttled asynchronous function.
        """

        @functools.wraps(func)
        async def wrapper(*args, **kwargs) -> Any:
            elapsed = _t() - self.last_called
            if elapsed < self.period:
                await asyncio.sleep(self.period - elapsed)
            self.last_called = _t()
            return await func(*args, **kwargs)

        return wrapper


def throttle(func: Callable[..., T], period: float) -> Callable[..., Callable[..., T]]:
    """
    Throttle function execution to limit the rate of calls.

    Args:
        func: The function to throttle.
        period: The minimum time interval between consecutive calls.

    Returns:
        The throttled function.
    """
    if not is_coroutine_func(func):
        func = force_async(func)
    throttle_instance = Throttle(period)

    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        await throttle_instance(func)(*args, **kwargs)
        return await func(*args, **kwargs)

    return wrapper



import json
from collections.abc import AsyncGenerator
from os import getenv
from typing import Any

import aiohttp
from pydantic import BaseModel, Field, field_validator

from .data_models import (
    AnthropicEndpointPathParam,
    AnthropicEndpointQueryParam,
    AnthropicEndpointRequestBody,
)
from .match_response import match_response


class AnthropicRequest(BaseModel):
    api_key: str = Field(description="API key for authentication", exclude=True)

    endpoint: str = Field(description="Endpoint for request")

    method: str = Field(description="HTTP method")

    content_type: str | None = Field(default=None, description="HTTP Content-Type")

    api_version: str = Field(default="2023-06-01", description="Anthropic API version")

    model_config = {
        "arbitrary_types_allowed": True,
        "extra": "allow",  # Allow extra attributes for mocking
    }

    @field_validator("api_key")
    @classmethod
    def get_id(cls, value: str) -> str:
        try:
            if api_key := getenv(value):
                return api_key
            else:
                return value
        except Exception:
            return value

    def get_endpoint(self, path_param: AnthropicEndpointPathParam = None):
        if path_param is None:
            return self.endpoint
        else:
            return self.endpoint.format(**path_param.model_dump())

    @property
    def base_url(self):
        return "https://api.anthropic.com/v1/"

    async def invoke(
        self,
        json_data: AnthropicEndpointRequestBody = None,
        params: AnthropicEndpointQueryParam = None,
        path_param: AnthropicEndpointPathParam = None,
        output_file: str = None,
        with_response_header: bool = False,
        parse_response: bool = True,
    ) -> Any:
        def get_headers():
            headers = {
                "x-api-key": self.api_key,
                "anthropic-version": self.api_version,
            }
            if self.content_type:
                headers["Content-Type"] = self.content_type
            return headers

        url = self.base_url + self.get_endpoint(path_param)
        headers = get_headers()
        json_data = json_data.model_dump(exclude_unset=True) if json_data else None
        params = params.model_dump(exclude_unset=True) if params else None

        async with aiohttp.ClientSession() as client:
            async with client.request(
                method=self.method,
                url=url,
                headers=headers,
                json=json_data,
                params=params,
            ) as response:
                if response.status != 200:
                    try:
                        error_text = await response.json()
                    except Exception:
                        error_text = await response.text()
                    raise aiohttp.ClientResponseError(
                        request_info=response.request_info,
                        history=response.history,
                        status=response.status,
                        message=f"{error_text}",
                        headers=response.headers,
                    )

                # handle stream in messages
                if json_data and json_data.get("stream"):
                    response_body = []
                    file_handle = None
                    if output_file:
                        try:
                            file_handle = open(output_file, "w")
                        except Exception as e:
                            raise ValueError(
                                f"Invalid to output the response "
                                f"to {output_file}. Error:{e}"
                            )

                    try:
                        async for chunk in response.content:
                            chunk_str = chunk.decode("utf-8")
                            if chunk_str.startswith("data: "):
                                chunk_str = chunk_str[6:]  # Remove "data: " prefix
                            if chunk_str.strip() == "[DONE]":
                                continue
                            try:
                                chunk_data = json.loads(chunk_str)
                                if file_handle:
                                    file_handle.write(json.dumps(chunk_data) + "\n")
                                response_body.append(chunk_data)
                            except json.JSONDecodeError:
                                continue
                    except Exception as e:
                        raise e
                    finally:
                        if file_handle:
                            file_handle.close()

                    if parse_response:
                        response_body = match_response(self, response_body)
                    if with_response_header:
                        return response_body, response.headers
                    else:
                        return response_body

                if output_file:
                    try:
                        with open(output_file, "wb") as f:
                            f.write(await response.read())
                    except Exception as e:
                        raise ValueError(
                            f"Invalid to output the response "
                            f"to {output_file}. Error:{e}"
                        )

                response_body = await response.json()

                if parse_response:
                    response_body = match_response(self, response_body)
                if with_response_header:
                    return response_body, response.headers
                else:
                    return response_body

    async def stream(
        self,
        json_data: AnthropicEndpointRequestBody,
        verbose: bool = True,
        output_file: str = None,
        with_response_header: bool = False,
    ) -> AsyncGenerator[Any, None]:
        if not getattr(json_data, "stream"):
            raise ValueError(
                "Request does not support stream. "
                "Only messages requests with stream=True are supported"
            )

        def get_headers():
            headers = {
                "x-api-key": self.api_key,
                "anthropic-version": self.api_version,
            }
            if self.content_type:
                headers["Content-Type"] = self.content_type
            return headers

        url = self.base_url + self.get_endpoint()
        headers = get_headers()
        json_data = json_data.model_dump(exclude_unset=True) if json_data else None

        try:
            async with aiohttp.ClientSession() as client:
                async with client.request(
                    method=self.method,
                    url=url,
                    headers=headers,
                    json=json_data,
                ) as response:
                    if response.status != 200:
                        try:
                            error_text = await response.json()
                        except Exception:
                            error_text = await response.text()
                        raise aiohttp.ClientResponseError(
                            request_info=response.request_info,
                            history=response.history,
                            status=response.status,
                            message=f"{error_text}",
                            headers=response.headers,
                        )

                    file_handle = None
                    if output_file:
                        try:
                            file_handle = open(output_file, "w")
                        except Exception as e:
                            raise ValueError(
                                f"Invalid to output the response "
                                f"to {output_file}. Error:{e}"
                            )

                    try:
                        async for chunk in response.content:
                            chunk_str = chunk.decode("utf-8")
                            if chunk_str.startswith("data: "):
                                chunk_str = chunk_str[6:]  # Remove "data: " prefix
                            if chunk_str.strip() == "[DONE]":
                                continue
                            try:
                                chunk_data = json.loads(chunk_str)
                                if file_handle:
                                    file_handle.write(json.dumps(chunk_data) + "\n")

                                # Print content when verbose is True
                                if verbose:
                                    if chunk_data.get("type") == "content_block_delta":
                                        if text := chunk_data.get("delta", {}).get(
                                            "text"
                                        ):
                                            print(text, end="", flush=True)

                                yield chunk_data
                            except json.JSONDecodeError:
                                continue

                        if with_response_header:
                            yield response.headers
                    except Exception as e:
                        raise e
                    finally:
                        if file_handle:
                            file_handle.close()
        except Exception as e:
            raise e

    def __repr__(self):
        return (
            f"AnthropicRequest(endpoint={self.endpoint}, method={self.method}, "
            f"content_type={self.content_type}, api_version={self.api_version})"
        )



from collections import deque
from collections.abc import Generator
from typing import Any, TypeVar

from lion.core.typing import ID, IDError

from .element import Element

T = TypeVar("T")


def to_list_type(value: Any, /) -> list[Any]:
    """Convert input to a list format"""
    if value is None:
        return []
    if isinstance(value, str):
        return [value] if ID.is_id(value) else []
    if isinstance(value, Element):
        return [value]
    if hasattr(value, "values") and callable(value.values):
        return list(value.values())
    if isinstance(value, list | tuple | set | deque | Generator):
        return list(value)
    return [value]


def validate_order(value: Any, /) -> list[str]:
    """Validate and standardize order representation"""

    try:
        result = []
        for item in to_list_type(value):
            if isinstance(item, str) and ID.is_id(item):
                result.append(item)
            elif isinstance(item, Element):
                result.append(item.ln_id)
            else:
                id_ = ID.get_id(item)
                if id_:
                    result.append(id_)
        return result
    except Exception as e:
        raise IDError("Must only contain valid Lion IDs.") from e


# File: autoos/container/util.py



from typing import Literal

from pydantic import BaseModel, Field


class Function(BaseModel):
    name: str = Field(description="The name of the function to call.")
    arguments: str = Field(description="The arguments to pass to the function.")


class ToolCall(BaseModel):
    id: str = Field(description="The ID of the tool call.")
    type: Literal["function"] = Field(description="The type of the tool call.")
    function: Function = Field(description="The function call details.")



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from lion.core.generic import Element, Log
from lion.core.typing import Any, Enum, NoReturn, PrivateAttr, override
from lion.settings import Settings, TimedFuncCallConfig


class EventStatus(str, Enum):
    """Status states for tracking action execution progress.

    Attributes:
        PENDING: Initial state before execution starts
        PROCESSING: Action is currently being executed
        COMPLETED: Action completed successfully
        FAILED: Action failed during execution
    """

    PENDING = "pending"
    PROCESSING = "processing"
    COMPLETED = "completed"
    FAILED = "failed"


class ObservableAction(Element):
    """Base class for actions that can be monitored and tracked.

    Provides functionality for tracking execution status, timing, and results
    of action invocations. Inherits from Element for serialization capabilities.

    Attributes:
        status (EventStatus): Current execution status of the action
        execution_time (float | None): Time taken to execute the action in seconds
        execution_response (Any): Result returned from the action execution
        execution_error (str | None): Error message if execution failed
        _timed_config (TimedFuncCallConfig): Configuration for timed execution
        _content_fields (list): Fields to include in content section of logs
    """

    status: EventStatus = EventStatus.PENDING
    execution_time: float | None = None
    execution_response: Any = None
    execution_error: str | None = None
    _timed_config: TimedFuncCallConfig | None = PrivateAttr(None)
    _content_fields: list = PrivateAttr(["execution_response"])

    @override
    def __init__(
        self, timed_config: dict | TimedFuncCallConfig | None, **kwargs: Any
    ) -> None:
        """Initialize an ObservableAction instance.

        Args:
            timed_config: Configuration for timed function execution.
                If None, uses default config from Settings.
                If dict, converts to TimedFuncCallConfig.
            **kwargs: Additional keyword arguments passed to TimedFuncCallConfig
                if timed_config is a dict.
        """
        super().__init__()
        if timed_config is None:
            self._timed_config = Settings.Config.TIMED_CALL

        else:
            if isinstance(timed_config, TimedFuncCallConfig):
                timed_config = timed_config.to_dict()
            if isinstance(timed_config, dict):
                timed_config = {**timed_config, **kwargs}
            timed_config = TimedFuncCallConfig(**timed_config)
            self._timed_config = timed_config

    def to_log(self) -> Log:
        """Convert the action to a log entry.

        Creates a structured log entry with execution details split into
        content and loginfo sections. Content includes execution results,
        while loginfo includes metadata and status information.

        Returns:
            Log: A log entry representing the action's execution state
                and results.
        """
        dict_ = self.to_dict()
        dict_["status"] = self.status.value
        content = {k: dict_[k] for k in self._content_fields if k in dict_}
        loginfo = {k: dict_[k] for k in dict_ if k not in self._content_fields}
        return Log(content=content, loginfo=loginfo)

    @classmethod
    def from_dict(cls, data: dict, /, **kwargs: Any) -> NoReturn:
        """Event cannot be re-created from dictionary.

        This method is intentionally not implemented as actions represent
        one-time events that should not be recreated from serialized state.

        Args:
            data: Dictionary of serialized action data
            **kwargs: Additional keyword arguments

        Raises:
            NotImplementedError: Always raised as actions cannot be recreated
        """
        raise NotImplementedError(
            "An event cannot be recreated. Once it's done, it's done."
        )


__all__ = ["ObservableAction"]



from .delete_fine_tuned_model import OpenAIDeleteFineTunedModelPathParam
from .retrieve_model import OpenAIRetrieveModelPathParam

__all__ = [
    "OpenAIRetrieveModelPathParam",
    "OpenAIDeleteFineTunedModelPathParam",
]



from typing import List, Optional

from pydantic import Field

from ..data_models import OpenAIEndpointPathParam, OpenAIEndpointRequestBody


class OpenAICompleteUploadPathParam(OpenAIEndpointPathParam):
    upload_id: str = Field(description="The ID of the Upload.")


class OpenAICompleteUploadRequestBody(OpenAIEndpointRequestBody):
    part_ids: list[str] = Field(description="The ordered list of Part IDs.")

    md5: str | None = Field(
        description="The optional md5 checksum for the file contents to verify "
        "if the bytes uploaded matches what you expect."
    )



"""
Copyright 2024 HaiyangLi

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import inspect
from enum import Enum
from typing import Any

from lion.core.typing import BaseModel, JsonValue
from lion.libs.parse import is_same_dtype
from lion.libs.string_similarity import string_similarity


def parse_to_representation(
    choices: Enum | dict | list | tuple | set,
) -> tuple[list[str], JsonValue]:
    """
    should use
    1. iterator of string | BaseModel
    2. dict[str, JsonValue | BaseModel]
    3. Enum[str, JsonValue | BaseModel]
    """

    if isinstance(choices, tuple | set | list):
        choices = list(choices)
        if is_same_dtype(choices, str):
            return choices, choices

    if isinstance(choices, list):
        if is_same_dtype(choices, BaseModel):
            choices = {i.__class__.__name__: i for i in choices}
        if all(inspect.isclass(i) and issubclass(i, BaseModel) for i in choices):
            choices = {i.__name__: i for i in choices}
    if isinstance(choices, type) and issubclass(choices, Enum):
        keys = [i.name for i in choices]
        contents = [get_choice_representation(i) for i in choices]
        return keys, contents

    if isinstance(choices, dict):
        keys = list(choices.keys())
        contents = list(choices.values())
        contents = [get_choice_representation(v) for k, v in choices.items()]
        return keys, contents

    if isinstance(choices, tuple | set | list):
        choices = list(choices)
        if is_same_dtype(choices, str):
            return choices, choices

    raise NotImplementedError


def get_choice_representation(choice: Any) -> str:

    if isinstance(choice, str):
        return choice

    if isinstance(choice, BaseModel):
        return f"{choice.__class__.__name__}:\n{choice.model_json_schema(indent=2)}"

    if isinstance(choice, Enum):
        return get_choice_representation(choice.value)


def parse_selection(selection_str: str, choices: Any):

    select_from = []

    if isinstance(choices, dict):
        select_from = list(choices.keys())

    if inspect.isclass(choices) and issubclass(choices, Enum):
        select_from = [choice.name for choice in choices]

    if isinstance(choices, list | tuple | set):
        if is_same_dtype(choices, BaseModel):
            select_from = [i.__class__.__name__ for i in choices]
        if is_same_dtype(choices, str):
            select_from = list(choices)
        if all(inspect.isclass(i) and issubclass(i, BaseModel) for i in choices):
            select_from = [i.__name__ for i in choices]

    if not select_from:
        raise ValueError("The values provided for choice is not valid")

    selected = string_similarity(selection_str, select_from, return_most_similar=True)

    if isinstance(choices, dict) and selected in choices:
        return choices[selected]

    if inspect.isclass(choices) and issubclass(choices, Enum):
        for i in choices:
            if i.name == selected:
                return i

    if isinstance(choices, list) and is_same_dtype(choices, str):
        if selected in choices:
            return selected

    return selected



import warnings
from pathlib import Path
from typing import Any, Dict, Optional, Union

import yaml
from dotenv import load_dotenv
from pydantic import BaseModel, ConfigDict, Field, field_serializer, model_validator

from lion.service.rate_limiter import RateLimiter, RateLimitError
from lion.service.service_util import invoke_retry
from lion.service.token_calculator import TiktokenCalculator

from .api_endpoints.data_models import (
    GroqChatCompletionRequest,
    GroqEndpointRequestBody,
)
from .api_endpoints.groq_request import GroqRequest
from .api_endpoints.response_utils import match_response

load_dotenv()
path = Path(__file__).parent

price_config_file_name = path / "groq_price_data.yaml"
max_output_token_file_name = path / "groq_max_output_token_data.yaml"
rate_limits_file_name = path / "groq_rate_limits.yaml"


class GroqModel(BaseModel):
    model: str = Field(description="ID of the model to use.")
    request_model: GroqRequest = Field(description="Making requests")
    rate_limiter: RateLimiter | None = Field(
        default=None, description="Rate Limiter to track usage"
    )
    text_token_calculator: TiktokenCalculator | None = Field(
        default=None, description="Token Calculator"
    )
    estimated_output_len: int = Field(
        default=0, description="Expected output len before making request"
    )

    model_config = ConfigDict(extra="forbid")

    @model_validator(mode="before")
    @classmethod
    def parse_input(cls, data: dict):
        if not isinstance(data, dict):
            raise ValueError("Invalid init param")

        # Parse request model
        request_model_params = {
            "api_key": data.pop("api_key", None),
            "endpoint": data.pop("endpoint", None),
            "method": data.pop("method", None),
            "content_type": data.pop("content_type", None),
        }
        data["request_model"] = GroqRequest(**request_model_params)

        # Load rate limits from YAML
        try:
            with open(rate_limits_file_name) as file:
                rate_limits = yaml.safe_load(file)
                model_name = data.get("model")
                model_limits = None

                if model_name in rate_limits.get("text_models", {}):
                    model_limits = rate_limits["text_models"][model_name]
                elif model_name in rate_limits.get("audio_models", {}):
                    model_limits = rate_limits["audio_models"][model_name]

                if model_limits:
                    rate_limiter_params = {}
                    if limit_tokens := data.pop("limit_tokens", None):
                        rate_limiter_params["limit_tokens"] = limit_tokens
                    elif "tokens_per_minute" in model_limits:
                        rate_limiter_params["limit_tokens"] = model_limits[
                            "tokens_per_minute"
                        ]

                    if limit_requests := data.pop("limit_requests", None):
                        rate_limiter_params["limit_requests"] = limit_requests
                    elif "requests_per_minute" in model_limits:
                        rate_limiter_params["limit_requests"] = model_limits[
                            "requests_per_minute"
                        ]

                    if rate_limiter_params:
                        data["rate_limiter"] = RateLimiter(**rate_limiter_params)
        except FileNotFoundError:
            warnings.warn(f"Rate limits file not found: {rate_limits_file_name}")

        # Initialize token calculator
        try:
            text_calc = TiktokenCalculator(encoding_name="cl100k_base")
            data["text_token_calculator"] = text_calc
        except Exception as e:
            warnings.warn(f"Failed to initialize token calculator: {str(e)}")

        return data

    @field_serializer("request_model")
    def serialize_request_model(self, value: GroqRequest):
        return value.model_dump(exclude_unset=True)

    @invoke_retry(max_retries=3, base_delay=1, max_delay=60)
    async def invoke(self, **kwargs):
        """Make a request to the Groq API."""
        # Extract request body from kwargs
        request_body = kwargs.get("request_body")
        if not request_body:
            raise ValueError("request_body is required")

        # Extract other parameters
        estimated_output_len = kwargs.get("estimated_output_len", 0)
        output_file = kwargs.get("output_file")
        parse_response = kwargs.get("parse_response", True)

        if request_model := getattr(request_body, "model", None):
            if request_model != self.model:
                raise ValueError(
                    f"Request model does not match. Model is {self.model}, but request is made for {request_model}."
                )

        # Check rate limits if enabled
        if self.rate_limiter:
            input_token_len = await self.get_input_token_len(request_body)
            if not self.verify_invoke_viability(
                input_tokens_len=input_token_len,
                estimated_output_len=estimated_output_len,
            ):
                raise RateLimitError("Rate limit reached")

        try:
            if getattr(request_body, "stream", False):
                return await self.stream(
                    request_body=request_body,
                    output_file=output_file,
                    parse_response=parse_response,
                )

            response_body, response_headers = await self.request_model.invoke(
                request_body=request_body,
                output_file=output_file,
                with_response_header=True,
                parse_response=False,
            )

            if response_body:
                # Update rate limits based on usage
                if self.rate_limiter:
                    if usage := response_body.get("usage"):
                        total_tokens = usage.get("total_tokens", 0)
                        self.rate_limiter.update_rate_limit(
                            response_headers.get("Date"), total_tokens
                        )
                    else:
                        self.rate_limiter.update_rate_limit(
                            response_headers.get("Date")
                        )

            if parse_response:
                return match_response(self.request_model, response_body)
            return response_body

        except Exception as e:
            raise e

    async def stream(
        self,
        request_body: GroqEndpointRequestBody,
        output_file: str | None = None,
        parse_response: bool = True,
    ):
        """Stream response from the Groq API."""
        response_chunks = []
        response_headers = None

        async for chunk in self.request_model.stream(
            request_body=request_body,
            output_file=output_file,
            with_response_header=True,
        ):
            if isinstance(chunk, dict):
                if "choices" in chunk or "usage" in chunk:
                    response_chunks.append(chunk)
                elif "headers" in chunk:
                    response_headers = chunk["headers"]

        # Update rate limits if usage information is available
        if self.rate_limiter and response_chunks:
            last_chunk = response_chunks[-1]
            if usage := last_chunk.get("usage"):
                total_tokens = usage.get("total_tokens", 0)
                if response_headers:
                    self.rate_limiter.update_rate_limit(
                        response_headers.get("Date"), total_tokens
                    )

        if parse_response:
            return match_response(self.request_model, response_chunks)
        return response_chunks

    async def get_input_token_len(
        self, request_body: GroqEndpointRequestBody | dict
    ) -> int:
        """Calculate the number of input tokens."""
        if not isinstance(request_body, (GroqEndpointRequestBody, dict)):
            return 0

        if isinstance(request_body, dict):
            request_body = GroqChatCompletionRequest(**request_body)

        if not self.text_token_calculator:
            warnings.warn(
                "Token calculator not available, using approximate token count"
            )
            # Approximate token count (1 token ≈ 4 characters)
            if isinstance(request_body, GroqChatCompletionRequest):
                messages_text = "\n".join(msg.content for msg in request_body.messages)
                return len(messages_text) // 4
            return 0

        total_tokens = 0
        if isinstance(request_body, GroqChatCompletionRequest):
            for message in request_body.messages:
                total_tokens += self.text_token_calculator.calculate(message.content)

        return total_tokens

    def verify_invoke_viability(
        self, input_tokens_len: int = 0, estimated_output_len: int = 0
    ) -> bool:
        """Check if the request can be made within rate limits."""
        if not self.rate_limiter:
            return True

        self.rate_limiter.release_tokens()

        estimated_output_len = (
            estimated_output_len
            if estimated_output_len != 0
            else self.estimated_output_len
        )

        if estimated_output_len == 0:
            try:
                with open(max_output_token_file_name) as file:
                    output_token_config = yaml.safe_load(file)
                    estimated_output_len = output_token_config.get(
                        self.model, 2048
                    )  # Default to 2048
                    self.estimated_output_len = estimated_output_len
            except FileNotFoundError:
                warnings.warn(
                    f"Max output token file not found: {max_output_token_file_name}"
                )
                # Use a conservative default
                estimated_output_len = 2048
                self.estimated_output_len = estimated_output_len

        return self.rate_limiter.check_availability(
            input_tokens_len, estimated_output_len
        )

    def estimate_text_price(
        self,
        input_text: str,
        estimated_num_of_output_tokens: int = 0,
    ) -> float:
        """Estimate the cost of processing text."""
        if not self.text_token_calculator:
            warnings.warn(
                "Token calculator not available, using approximate token count"
            )
            # Approximate token count (1 token ≈ 4 characters)
            num_of_input_tokens = len(input_text) // 4
        else:
            num_of_input_tokens = self.text_token_calculator.calculate(input_text)

        try:
            with open(price_config_file_name) as file:
                price_config = yaml.safe_load(file)
        except FileNotFoundError:
            raise ValueError(f"Price config file not found: {price_config_file_name}")

        if self.model not in price_config.get("model", {}):
            raise ValueError(f"No price information found for model: {self.model}")

        model_price_info = price_config["model"][self.model]
        estimated_price = (
            model_price_info["input_tokens"] * num_of_input_tokens
            + model_price_info["output_tokens"] * estimated_num_of_output_tokens
        )

        return estimated_price



from collections.abc import Iterable, Mapping
from typing import Any, overload

from pydantic import BaseModel
from pydantic_core import PydanticUndefinedType

from lion.libs.constants import UndefinedType


@overload
def to_list(input_: None | UndefinedType | PydanticUndefinedType, /) -> list: ...


@overload
def to_list(
    input_: str | bytes | bytearray, /, use_values: bool = False
) -> list[str | int]: ...


@overload
def to_list(input_: Mapping, /, use_values: bool = False) -> list[Any]: ...


@overload
def to_list(
    input_: Any,
    /,
    *,
    flatten: bool = False,
    dropna: bool = False,
    unique: bool = False,
) -> list: ...


def to_list(
    input_: Any,
    /,
    *,
    flatten: bool = False,
    dropna: bool = False,
    unique: bool = False,
    use_values: bool = False,
) -> list:
    """Convert various input types to a list.

    Handles different input types and converts them to a list, with options
    for flattening nested structures and removing None values.

    Args:
        input_: The input to be converted to a list.
        flatten: If True, flattens nested list structures.
        dropna: If True, removes None values from the result.
        unique: If True, returns only unique values (requires flatten=True).
        use_values: If True, uses .values() for dict-like inputs.

    Returns:
        A list derived from the input, processed as specified.

    Raises:
        ValueError: If unique=True and flatten=False.

    Examples:
        >>> to_list(1)
        [1]
        >>> to_list([1, [2, 3]], flatten=True)
        [1, 2, 3]
        >>> to_list([1, None, 2], dropna=True)
        [1, 2]
        >>> to_list({'a': 1, 'b': 2}, use_values=True)
        [1, 2]
    """
    if unique and not flatten:
        raise ValueError("unique=True requires flatten=True")

    lst_ = _to_list_type(input_, use_values=use_values)

    if any((flatten, dropna)):
        lst_ = _process_list(
            lst=lst_,
            flatten=flatten,
            dropna=dropna,
        )

    return list(set(lst_)) if unique else lst_


def _undefined_to_list(input_: None | UndefinedType | PydanticUndefinedType, /) -> list:
    return []


def _str_to_list(
    input_: str | bytes | bytearray, /, use_values: bool = False
) -> list[str | int]:
    if use_values:
        return list(input_)
    return [input_]


def _dict_to_list(input_: Mapping, /, use_values: bool = False) -> list[Any]:
    if use_values:
        return list(input_.values())
    return [input_]


def _to_list_type(input_: Any, /, use_values: bool = False) -> Any | None:

    if isinstance(input_, BaseModel):
        return [input_]

    if use_values and hasattr(input_, "values"):
        return list(input_.values())

    if isinstance(input_, list):
        return input_

    if isinstance(input_, type(None) | UndefinedType | PydanticUndefinedType):
        return _undefined_to_list(input_)

    if isinstance(input_, str | bytes | bytearray):
        return _str_to_list(input_, use_values=use_values)

    if isinstance(input_, dict):
        return _dict_to_list(input_, use_values=use_values)

    if isinstance(input_, Iterable):
        return list(input_)

    return [input_]


def _process_list(lst: list[Any], flatten: bool, dropna: bool) -> list[Any]:
    """Process a list by optionally flattening and removing None values.

    Args:
        lst: The list to process.
        flatten: If True, flattens nested list structures.
        dropna: If True, removes None values.

    Returns:
        The processed list.
    """
    result = []
    for item in lst:
        if isinstance(item, Iterable) and not isinstance(
            item, (str, bytes, bytearray, Mapping)
        ):
            if flatten:
                result.extend(
                    _process_list(
                        lst=list(item),
                        flatten=flatten,
                        dropna=dropna,
                    )
                )
            else:
                result.append(
                    _process_list(
                        lst=list(item),
                        flatten=flatten,
                        dropna=dropna,
                    )
                )
        elif not dropna or item is not None:
            result.append(item)

    return result



from pydantic import Field

from ..data_models import OllamaEndpointRequestBody


class OllamaCopyModelRequestBody(OllamaEndpointRequestBody):
    source: str = Field(description="Name of the source existing model")

    destination: str = Field(description="Name of the copy model")



from typing import Optional

from pydantic import Field

from ..data_models import OllamaEndpointRequestBody, OllamaEndpointResponseBody


class OllamaCreateModelRequestBody(OllamaEndpointRequestBody):
    name: str = Field(description="name of the model to create")

    modelfile: str | None = Field(None, description="Contents of the Modelfile")

    stream: bool | None = Field(
        True,
        description="if 'false' the response will be returned as a single response object, "
        "rather than a stream of objects",
    )

    path: str | None = Field(None, description="path to the Modelfile")


class OllamaCreateModelResponseBody(OllamaEndpointResponseBody):
    status: str = Field(None)



from typing import List, Literal

from pydantic import BaseModel, ConfigDict, Field

from ..data_models import OpenAIEndpointResponseBody


class EmbeddingObject(BaseModel):
    index: int = Field(
        description="The index of the embedding in the list of embeddings."
    )

    embedding: list[float] = Field(
        description=(
            "The embedding vector, which is a list of floats. The length of "
            "vector depends on the model as listed in the embedding guide."
        ),
    )

    object: Literal["embedding"] = Field(
        "embedding",
        description='The object type, which is always "embedding".',
    )


class Usage(BaseModel):
    prompt_tokens: int = Field(description="Number of tokens in the prompt.")

    total_tokens: int = Field(
        description="Total number of tokens used in "
        "the request (prompt + completion)"
    )


class OpenAIEmbeddingResponseBody(OpenAIEndpointResponseBody):
    object: Literal["list"] = Field(
        description='The object type, which is always "list" for embeddings.'
    )
    data: list[EmbeddingObject] = Field(description="The list of embedding objects.")
    model: str = Field(
        description="The ID of the model used to generate the embeddings."
    )
    usage: Usage = Field(description="The usage statistics for the request.")

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "object": "list",
                "data": [
                    {
                        "index": 0,
                        "embedding": [
                            0.0023064255,
                            -0.009327292,
                            ...,
                            -0.0028842222,
                        ],
                        "object": "embedding",
                    }
                ],
                "model": "text-embedding-ada-002",
                "usage": {"prompt_tokens": 8, "total_tokens": 8},
            }
        }
    )



from pydantic import BaseModel, ConfigDict


class OpenAIEndpointRequestBody(BaseModel):
    model_config = ConfigDict(
        extra="forbid", use_enum_values=True, validate_assignment=True
    )


class OpenAIEndpointResponseBody(BaseModel):
    model_config = ConfigDict(use_enum_values=True, validate_assignment=True)


class OpenAIEndpointQueryParam(BaseModel):
    model_config = ConfigDict(
        extra="forbid", use_enum_values=True, validate_assignment=True
    )


class OpenAIEndpointPathParam(BaseModel):
    model_config = ConfigDict(
        extra="forbid", use_enum_values=True, validate_assignment=True
    )



from .cancel_jobs import OpenAICancelFineTuningPathParam
from .create_jobs import (
    Hyperparameters,
    Integration,
    OpenAICreateFineTuningJobRequestBody,
    Wandb,
)
from .list_fine_tuning_checkpoints import (
    OpenAIListFineTuningCheckpointsPathParam,
    OpenAIListFineTuningCheckpointsQueryParam,
)
from .list_fine_tuning_events import (
    OpenAIListFineTuningEventsPathParam,
    OpenAIListFineTuningEventsQueryParam,
)
from .list_fine_tuning_jobs import OpenAIListFineTuningJobsQueryParam
from .retrieve_jobs import OpenAIRetrieveFineTuningJobPathParam
from .training_format import (
    OpenAIChatModelTrainingFormat,
    OpenAICompletionsModelTrainingFormat,
)

__all__ = [
    "OpenAICreateFineTuningJobRequestBody",
    "Hyperparameters",
    "Integration",
    "Wandb",
    "OpenAIListFineTuningJobsQueryParam",
    "OpenAIListFineTuningEventsPathParam",
    "OpenAIListFineTuningEventsQueryParam",
    "OpenAIListFineTuningCheckpointsPathParam",
    "OpenAIListFineTuningCheckpointsQueryParam",
    "OpenAIRetrieveFineTuningJobPathParam",
    "OpenAICancelFineTuningPathParam",
    "OpenAIChatModelTrainingFormat",
    "OpenAICompletionsModelTrainingFormat",
]



PLAN_PROMPT = """
Develop a high-level plan with {num_steps} distinct steps. Each step should:
1. Represent a major milestone or phase
2. Be logically sequenced for dependencies
3. Be clearly distinct from other steps
4. Have measurable completion criteria
5. Be suitable for further decomposition
"""

EXPANSION_PROMPT = """
Break down a high-level plan into detailed concrete executable actions. Each step should:
- Ensure actions are atomic and verifiable
- Include necessary context and preconditions
- Specify expected outcomes and validations
- Maintain sequential dependencies
- Be self-contained with clear scope
- Include all required context/parameters
- Have unambiguous success criteria
- Specify error handling approach
- Define expected outputs
"""



import warnings

from lion.service.service import Service
from lion.service.service_match_util import (
    match_parameters,
    match_service,
    match_task_method,
)


class iModel:
    def __init__(
        self,
        provider: str | Service,
        task: str = "chat",
        model: str = None,
        api_key: str = None,
        api_key_schema: str = None,
        interval_tokens: int = None,
        interval_requests: int = None,
        **kwargs,
    ):
        if api_key is not None:
            api_key = api_key
        elif api_key_schema is not None:
            api_key = api_key_schema

        # Deprecation
        if "config" in kwargs:
            warnings.warn(
                "'config' is deprecated in 'iModel'. Please pass configurations directly as keyword arguments. "
                "If 'config' is a parameter name for further request settings, it will still be passed along.",
                DeprecationWarning,
                stacklevel=2,
            )
        if "provider_schema" in kwargs:
            warnings.warn(
                "'provider_schema' is deprecated in 'iModel'. Please refer to the corresponding service's "
                "Pydantic models for the schema structure. "
                "If 'provider_schema' is a parameter name for further request settings, it will still be passed along.",
                DeprecationWarning,
                stacklevel=2,
            )
        if "endpoint" in kwargs:
            warnings.warn(
                "'endpoint' is deprecated in 'iModel'. Please specify 'task' instead. "
                "If 'endpoint' is a parameter name for further request settings, it will still be passed along.",
                DeprecationWarning,
                stacklevel=2,
            )
            task = kwargs["endpoint"]
        if "token_encoding_name" in kwargs:
            warnings.warn(
                "'token_encoding_name' is deprecated in 'iModel' as it is now automatically detected. "
                "To explicitly set the encoding method, please refer to TokenCalculator.",
                DeprecationWarning,
                stacklevel=2,
            )
        if "interval" in kwargs:
            warnings.warn(
                "'interval' is deprecated in 'iModel' as the unit is now set per minute. "
                "If 'interval' is a parameter name for further request settings, it will still be passed along.",
                DeprecationWarning,
                stacklevel=2,
            )
        if "service" in kwargs:
            warnings.warn(
                "'service' parameter is deprecated in 'iModel'. Please use 'provider' instead. For more "
                "details about service configurations, refer to 'lion.service' or the corresponding service's package. "
                "If 'service' is a parameter name for further request settings, it will still be passed along.",
                DeprecationWarning,
                stacklevel=2,
            )
        if "allowed_parameters" in kwargs:
            warnings.warn(
                "'allowed_parameters' is deprecated in 'iModel'. If 'allowed_parameters' is a parameter "
                "name for further request settings, it will still be passed along.",
                DeprecationWarning,
                stacklevel=2,
            )
        if "device" in kwargs:
            warnings.warn(
                "'device' is deprecated in 'iModel'. "
                "If 'device' is a parameter name for further request settings, it will still be passed along.",
                DeprecationWarning,
                stacklevel=2,
            )
        if "costs" in kwargs:
            warnings.warn(
                "'cost' is deprecated in 'iModel'. "
                "If 'costs' is a parameter name for further request settings, it will still be passed along.",
                DeprecationWarning,
                stacklevel=2,
            )

        if isinstance(provider, str):
            self.service = match_service(provider, api_key=api_key, **kwargs)
        elif isinstance(provider, Service):
            self.service = provider
            if api_key:
                warnings.warn(
                    "A Service instance was provided along with api key info."
                    "The the separately provided api_key or api_key_schema will be ignored."
                )
        else:
            raise ValueError(
                "Invalid provider. Please provide a valid provider name or valid service object."
            )

        task_method_list = match_task_method(task, self.service)
        if len(task_method_list) == 0:
            raise ValueError(
                "No matching task found. "
                "Please refer to the service provider's API to provide a valid task."
            )
        if len(task_method_list) > 1:
            raise ValueError(
                f"Multiple possible tasks found. Please specify: {task_method_list}"
            )
        self.task = task_method_list[0]
        task_method = getattr(self.service, task_method_list[0])
        task_params = match_parameters(
            task_method, model, interval_tokens, interval_requests
        )
        try:
            self.request_model = task_method(**task_params)
        except Exception as e:
            raise ValueError(
                f"{self.task} requires the following to be provided as input: {e}."
            )

        self.model = model
        self.configs = kwargs

        self.data_model = self.service.match_data_model(self.task)

    def parse_to_data_model(self, **kwargs):
        if kwargs.get("model") and self.model:
            if kwargs.get("model") != self.model:
                raise ValueError(
                    f"Models are inconsistent. This iModel is for {self.model}"
                )

        output = {}
        for invoke_param, data_model in self.data_model.items():
            data_model_dict = {}
            if "model" in data_model.model_fields:
                data_model_dict["model"] = self.model
            for key in self.configs:
                if key in data_model.model_fields:
                    data_model_dict[key] = self.configs[key]
            for key in kwargs:
                if key in data_model.model_fields:
                    data_model_dict[key] = kwargs[key]
            output[invoke_param] = data_model(**data_model_dict)
        return output

    async def invoke(self, **kwargs):
        return await self.request_model.invoke(**kwargs)

    def list_tasks(self):
        return self.service.list_tasks()


__all__ = ["iModel"]



from lion.core.models.schema_model import SchemaModel


class LionIDConfig(SchemaModel):
    n: int
    random_hyphen: bool
    num_hyphens: int
    hyphen_start_index: int
    hyphen_end_index: int
    prefix: str = "ln"
    postfix: str = ""



from enum import Enum
from typing import IO, Literal, Optional

from pydantic import ConfigDict, Field, field_validator

from ..data_models import OpenAIEndpointRequestBody


class Size(str, Enum):
    S256 = "256x256"
    S512 = "512x512"
    S1024 = "1024x1024"


class OpenAIImageEditRequestBody(OpenAIEndpointRequestBody):
    image: str | IO = Field(
        description="The image to edit. Must be a valid PNG file, less than 4MB, and square.",
    )

    prompt: str = Field(
        description="A text description of the desired image(s).",
        max_length=1000,
    )

    mask: str | IO | None = Field(
        None,
        description="An additional image whose fully transparent areas indicate where image should be edited.",
    )

    model: Literal["dall-e-2"] | None = Field(
        "dall-e-2", description="The model to use for image generation."
    )

    n: int | None = Field(
        1, description="The number of images to generate.", ge=1, le=10
    )

    size: Size | None = Field(
        "1024x1024", description="The size of the generated images."
    )

    response_format: Literal["url", "b64_json"] | None = Field(
        "url",
        description="The format in which the generated images are returned.",
    )

    user: str | None = Field(
        None,
        description="A unique identifier representing your end-user, which can help OpenAI to monitor and detect abuse.",
    )

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("image", "mask")
    @classmethod
    def get_file_object(cls, value):
        if isinstance(value, str):
            value = open(value, "rb")
        return value

    def __del__(self):
        # Ensure files are closed when the object is destroyed
        image_obj = self.__dict__.get("image")
        if image_obj and not image_obj.closed:
            image_obj.close()

        mask_obj = self.__dict__.get("mask")
        if mask_obj and not mask_obj.closed:
            mask_obj.close()



import json
from collections.abc import AsyncGenerator
from typing import Any, Dict, Optional, Tuple, Union

import aiohttp
from pydantic import BaseModel, ConfigDict, Field

from lion.integrations.perplexity_.api_endpoints.data_models import (
    PerplexityEndpointRequestBody,
)


class PerplexityRequest(BaseModel):
    """Handler for making requests to the Perplexity API."""

    api_key: str = Field(description="API key for authentication", exclude=True)
    endpoint: str = Field(description="Endpoint for request")
    method: str = Field(description="HTTP method")
    content_type: str | None = "application/json"
    base_url: str = "https://api.perplexity.ai"

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="allow",  # Allow extra attributes for mocking in tests
    )

    async def invoke(
        self,
        json_data: None | (dict[str, Any] | PerplexityEndpointRequestBody) = None,
        form_data: BaseModel | None = None,
        output_file: str | None = None,
        with_response_header: bool = False,
        parse_response: bool = True,
    ) -> dict[str, Any] | tuple[dict[str, Any], dict[str, str]] | bytes | None:
        """Make a request to the Perplexity API."""
        url = f"{self.base_url}/{self.endpoint}"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
        }

        if self.content_type:
            headers["Content-Type"] = self.content_type

        if isinstance(json_data, PerplexityEndpointRequestBody):
            json_data = json_data.model_dump(exclude_unset=True)

        async with aiohttp.ClientSession() as client:
            response = await client.request(
                method=self.method,
                url=url,
                headers=headers,
                json=json_data,
                data=form_data.model_dump() if form_data else None,
            )
            async with response:
                if response.status != 200:
                    try:
                        error_body = await response.json()
                        error_msg = error_body.get("error", {}).get(
                            "message", str(error_body)
                        )
                    except Exception:
                        error_msg = await response.text()
                    raise Exception(
                        f"API request failed with status {response.status}: {error_msg}"
                    )

                if output_file:
                    with open(output_file, "wb") as f:
                        async for chunk in response.content.iter_chunked(8192):
                            f.write(chunk)
                    return None

                if parse_response:
                    response_body = await response.json()
                else:
                    response_body = await response.text()
                    try:
                        response_body = json.loads(response_body)
                    except json.JSONDecodeError:
                        pass

                if with_response_header:
                    headers = {k.lower(): v for k, v in response.headers.items()}
                    return response_body, headers
                return response_body

    async def stream(
        self,
        json_data: None | (dict[str, Any] | PerplexityEndpointRequestBody) = None,
        output_file: str | None = None,
        with_response_header: bool = False,
        verbose: bool = True,
    ) -> AsyncGenerator[dict[str, Any], None]:
        """Stream responses from the Perplexity API."""
        if isinstance(json_data, PerplexityEndpointRequestBody):
            json_data = json_data.model_dump(exclude_unset=True)

        if not json_data.get("stream", False):
            json_data["stream"] = True

        url = f"{self.base_url}/{self.endpoint}"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
            "Accept": "text/event-stream",
        }

        async with aiohttp.ClientSession() as client:
            response = await client.post(url, headers=headers, json=json_data)
            async with response:
                if response.status != 200:
                    try:
                        error_body = await response.json()
                        error_msg = error_body.get("error", {}).get(
                            "message", str(error_body)
                        )
                    except Exception:
                        error_msg = await response.text()
                    raise Exception(
                        f"API request failed with status {response.status}: {error_msg}"
                    )

                if with_response_header:
                    headers = {k.lower(): v for k, v in response.headers.items()}
                    yield {"headers": headers}

                file_handle = None
                if output_file:
                    try:
                        file_handle = open(output_file, "w")
                    except Exception as e:
                        raise ValueError(
                            f"Failed to open output file {output_file}: {e}"
                        )

                try:
                    async for chunk in response.content:
                        if chunk:
                            chunk_str = chunk.decode("utf-8").strip()
                            if chunk_str.startswith("data: "):
                                chunk_str = chunk_str[6:]  # Remove "data: " prefix
                            try:
                                chunk_data = json.loads(chunk_str)
                                if file_handle:
                                    file_handle.write(json.dumps(chunk_data) + "\n")
                                if verbose and "choices" in chunk_data:
                                    content = chunk_data["choices"][0]["delta"][
                                        "content"
                                    ]
                                    print(content, end="", flush=True)
                                yield chunk_data
                            except json.JSONDecodeError:
                                continue
                finally:
                    if file_handle:
                        file_handle.close()



import json

import aiohttp
from pydantic import BaseModel, Field

from .data_models import OllamaEndpointRequestBody
from .match_response import match_response


class OllamaRequest(BaseModel):
    endpoint: str = Field(description="Endpoint for request")

    method: str = Field(description="HTTP method")

    @property
    def base_url(self):
        return "http://localhost:11434/api/"

    async def invoke(
        self,
        json_data: OllamaEndpointRequestBody = None,
        output_file: str = None,
        with_response_header: bool = False,
        parse_response: bool = True,
    ):
        url = self.base_url + self.endpoint
        json_data = json_data.model_dump(exclude_unset=True) if json_data else None

        async with aiohttp.ClientSession() as client:
            async with client.request(
                method=self.method, url=url, json=json_data
            ) as response:
                if response.status != 200:
                    try:
                        error_text = await response.json()
                    except:
                        error_text = await response.text()
                    raise aiohttp.ClientResponseError(
                        request_info=response.request_info,
                        history=response.history,
                        status=response.status,
                        message=f"HTTP Error {response.status}. Response Body: {error_text}",
                        headers=response.headers,
                    )

                # handle stream
                if response.headers.get("Content-Type") == "application/x-ndjson":
                    response_body = []
                    file_handle = None
                    if output_file:
                        try:
                            file_handle = open(output_file, "w")
                        except Exception as e:
                            raise ValueError(
                                f"Invalid to output the response to {output_file}. Error:{e}"
                            )
                    try:
                        async for chunk in response.content:
                            chunk_str = chunk.decode("utf-8")
                            chunk_str = chunk_str.strip()
                            response_body.append(json.loads(chunk_str))
                            if "status" in json.loads(chunk_str):
                                print(json.loads(chunk_str))
                            if file_handle:
                                file_handle.write(chunk_str + "\n")

                    finally:
                        if file_handle:
                            file_handle.close()

                    if parse_response:
                        response_body = match_response(self, response_body)
                    if with_response_header:
                        return response_body, response.headers
                    else:
                        return response_body
                else:
                    try:
                        response_body = await response.json()
                    except:
                        response_body = await response.text()

                    if output_file:
                        try:
                            with open(output_file, "wb") as f:
                                f.write(await response.read())
                        except Exception as e:
                            raise ValueError(
                                f"Invalid to output the response to {output_file}. Error:{e}"
                            )

                    if parse_response:
                        response_body = match_response(self, response_body)
                    if with_response_header:
                        return response_body, response.headers
                    else:
                        return response_body

    async def stream(
        self,
        json_data: OllamaEndpointRequestBody,
        verbose: bool = True,
        output_file: str = None,
        with_response_header: bool = False,
    ):

        if not getattr(json_data, "stream", None):
            raise ValueError(
                "Request does not support stream or is not in stream mode. "
                "Only requests with stream=True are supported."
            )

        url = self.base_url + self.endpoint
        json_data = json_data.model_dump(exclude_unset=True)

        async with aiohttp.ClientSession() as client:
            async with client.request(
                method=self.method, url=url, json=json_data
            ) as response:
                if response.status != 200:
                    try:
                        error_text = await response.json()
                    except:
                        error_text = await response.text()
                    raise aiohttp.ClientResponseError(
                        request_info=response.request_info,
                        history=response.history,
                        status=response.status,
                        message=f"{error_text}",
                        headers=response.headers,
                    )

                file_handle = None

                if output_file:
                    try:
                        file_handle = open(output_file, "w")
                    except Exception as e:
                        raise ValueError(
                            f"Invalid to output the response to {output_file}. Error:{e}"
                        )

                try:
                    async for chunk in response.content:
                        chunk_str = chunk.decode("utf-8")
                        chunk_str = chunk_str.strip()
                        if file_handle:
                            file_handle.write(chunk_str + "\n")

                        chunk_dict = json.loads(chunk_str)
                        if verbose:
                            if chunk_dict.get("response"):
                                print(chunk_dict.get("response"), end="", flush=True)
                            if message := chunk_dict.get("message"):
                                print(message.get("content"), end="", flush=True)

                        yield chunk_dict

                    if with_response_header:
                        yield response.headers

                finally:
                    if file_handle:
                        file_handle.close()



from pydantic import Field

from ..data_models import OpenAIEndpointPathParam


class OpenAICancelFineTuningPathParam(OpenAIEndpointPathParam):
    fine_tuning_job_id: str = Field(
        description="The ID of the fine-tuning job to get events for."
    )



"""String similarity calculation and matching functionality."""

from .algorithms import (
    SIMILARITY_ALGO_MAP,
    cosine_similarity,
    hamming_similarity,
    jaro_distance,
    jaro_winkler_similarity,
    levenshtein_distance,
    levenshtein_similarity,
    sequence_matcher_similarity,
)
from .matcher import MatchResult, string_similarity
from .types import SIMILARITY_TYPE, SimilarityFunc

__all__ = [
    # Main functionality
    "string_similarity",
    "MatchResult",
    # Algorithms
    "cosine_similarity",
    "hamming_similarity",
    "jaro_distance",
    "jaro_winkler_similarity",
    "levenshtein_distance",
    "levenshtein_similarity",
    "sequence_matcher_similarity",
    # Types and constants
    "SIMILARITY_TYPE",
    "SimilarityFunc",
    "SIMILARITY_ALGO_MAP",
]



from typing import List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator

from lion.integrations.perplexity_.api_endpoints.data_models import (
    Citation,
    PerplexityEndpointResponseBody,
    Usage,
)


class Message(BaseModel):
    """A message in the chat completion response."""

    role: str = Field(description="The role of the message author.")
    content: str = Field(description="The content of the message.")

    model_config = ConfigDict(extra="forbid")

    @field_validator("role")
    @classmethod
    def validate_role(cls, role: str) -> str:
        valid_roles = ["assistant"]  # Response messages are always from assistant
        if role not in valid_roles:
            raise ValueError(f"Role must be one of {valid_roles}")
        return role

    @field_validator("content")
    @classmethod
    def validate_content(cls, content: str) -> str:
        if not content or not content.strip():
            raise ValueError("Content cannot be empty")
        return content


class Choice(BaseModel):
    """A completion choice returned by the model."""

    index: int = Field(description="The index of this choice in the list of choices.")
    message: Message = Field(description="The message generated by the model.")
    finish_reason: str | None = Field(
        description="The reason the model stopped generating tokens. Either 'stop' or 'length'."
    )

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    @field_validator("finish_reason")
    @classmethod
    def validate_finish_reason(cls, finish_reason: str | None) -> str | None:
        if finish_reason is not None:
            valid_reasons = ["stop", "length"]
            if finish_reason not in valid_reasons:
                raise ValueError(f"Finish reason must be one of {valid_reasons}")
        return finish_reason


class RelatedQuestion(BaseModel):
    """A related question suggested by the model."""

    text: str = Field(description="The text of the related question.")

    model_config = ConfigDict(extra="forbid")

    @field_validator("text")
    @classmethod
    def validate_text(cls, text: str) -> str:
        if not text or not text.strip():
            raise ValueError("Text cannot be empty")
        return text


class PerplexityChatCompletionResponseBody(PerplexityEndpointResponseBody):
    """Response body for chat completion requests."""

    id: str = Field(description="A unique identifier for this completion.")
    model: str = Field(description="The model used for completion.")
    object: str = Field(description="The object type, always 'chat.completion'.")
    created: int = Field(
        description="Unix timestamp of when the completion was created."
    )
    choices: list[Choice] = Field(description="The list of completion choices.")
    usage: Usage = Field(description="Token usage information for this request.")
    citations: list[Citation | dict | str] | None = Field(
        default=None,
        description="Citations for the generated answer. Can be Citation objects, dictionaries, or URL strings.",
    )
    related_questions: list[RelatedQuestion] | None = Field(
        default=None, description="Related questions suggested by the model."
    )

    model_config = ConfigDict(extra="forbid")

    @field_validator("object")
    @classmethod
    def validate_object(cls, object_type: str) -> str:
        if object_type != "chat.completion":
            raise ValueError("Object type must be 'chat.completion'")
        return object_type

    @field_validator("choices")
    @classmethod
    def validate_choices(cls, choices: list[Choice]) -> list[Choice]:
        if not choices:
            raise ValueError("At least one choice is required")
        return choices

    @field_validator("citations")
    @classmethod
    def validate_citations(
        cls, citations: list[Citation | dict | str] | None
    ) -> list[Citation | dict | str] | None:
        if citations is not None:
            if len(citations) > 10:  # Arbitrary limit for safety
                raise ValueError("Too many citations")
            # Convert string citations to Citation objects
            processed_citations = []
            for citation in citations:
                if isinstance(citation, str):
                    processed_citations.append({"url": citation})
                else:
                    processed_citations.append(citation)
            return processed_citations
        return citations

    @field_validator("related_questions")
    @classmethod
    def validate_related_questions(
        cls, questions: list[RelatedQuestion] | None
    ) -> list[RelatedQuestion] | None:
        if questions is not None and len(questions) > 10:  # Arbitrary limit for safety
            raise ValueError("Too many related questions")
        return questions



from typing import List, Optional

from pydantic import BaseModel, Field

from ..data_models import OpenAIEndpointRequestBody


class Image(BaseModel):
    b64_json: str | None = Field(
        None,
        description="The base64-encoded JSON of the generated image, if response_format is b64_json.",
    )

    url: str | None = Field(
        None,
        description="The URL of the generated image, if response_format is url (default).",
    )

    revised_prompt: str | None = Field(
        None,
        description="The prompt that was used to generate the image, if there was any revision to the prompt.",
    )


class OpenAIImageResponseBody(OpenAIEndpointRequestBody):
    created: int = Field(
        description="The Unix timestamp of when the image was created."
    )

    data: list[Image] = Field(description="The list of generated images.")



import json
import os

import litellm
from dotenv import load_dotenv

litellm.drop_params = True
load_dotenv()


RESERVED_PARAMS = [
    "invoke_action",
    "instruction",
    "clear_messages",
]


class LiteiModel:

    def __init__(self, **kwargs):
        if "api_key" in kwargs:
            try:
                api_key1 = os.getenv(kwargs["api_key"], None)
                if api_key1:
                    # Store the original env var name as schema
                    self.api_key_schema = kwargs["api_key"]
                    # Store the resolved value for actual use
                    kwargs["api_key"] = api_key1
            except Exception:
                pass
        self.kwargs = kwargs
        self.acompletion = litellm.acompletion

    def to_dict(self) -> dict:
        dict_ = {k: v for k, v in self.kwargs.items() if k not in RESERVED_PARAMS}
        # If we have an api_key_schema, use that instead of the resolved value
        if hasattr(self, "api_key_schema"):
            dict_["api_key"] = self.api_key_schema
        return dict_

    @classmethod
    def from_dict(cls, data: dict) -> "LiteiModel":
        return cls(**data)

    async def invoke(self, **kwargs):
        config = {**self.kwargs, **kwargs}
        for i in RESERVED_PARAMS:
            config.pop(i, None)

        return await self.acompletion(**config)

    def __hash__(self):
        # Convert kwargs to a hashable format by serializing unhashable types
        hashable_items = []
        for k, v in self.kwargs.items():
            if isinstance(v, (list, dict)):
                # Convert unhashable types to JSON string for hashing
                v = json.dumps(v, sort_keys=True)
            elif not isinstance(v, (str, int, float, bool, type(None))):
                # Convert other unhashable types to string representation
                v = str(v)
            hashable_items.append((k, v))
        return hash(frozenset(hashable_items))



from enum import Enum

from pydantic import Field

from ..data_models import OpenAIEndpointResponseBody


class Purpose(str, Enum):
    ASSISTANTS = "assistants"
    ASSISTANTS_OUTPUT = "assistants_output"
    VISION = "vision"
    BATCH = "batch"
    BATCH_OUTPUT = "batch_output"
    FINE_TUNE = "fine-tune"
    FINE_TUNE_RESULTS = "fine-tune-results"


class Status(str, Enum):
    UPLOADED = "uploaded"
    PROCESSED = "processed"
    ERROR = "error"


class OpenAIFileResponseBody(OpenAIEndpointResponseBody):
    id: str = Field(
        description="The file identifier, which can be "
        "referenced in the API endpoints.",
    )

    bytes: int = Field(description="The size of the file, in bytes.")

    created_at: int = Field(
        description="The Unix timestamp (in seconds) for " "when the file was created.",
    )

    filename: str = Field(description="The name of the file.")

    object: str = Field(description="The object type, which is always file.")

    purpose: Purpose = Field(description="The intended purpose of the file.")

    status: Status = Field(
        None,
        description="The current status of the file, which can be "
        "either uploaded, processed, or error.",
        deprecated=True,
    )

    status_details: str | None = Field(
        None,
        description="For details on why a fine-tuning training "
        "file failed validation, "
        "see the error field on fine_tuning.job.",
        deprecated=True,
    )



from abc import abstractmethod

import tiktoken
from pydantic import BaseModel, Field, field_validator


class TokenCalculator(BaseModel):

    @abstractmethod
    def calculate(self, *args, **kwargs): ...


class TiktokenCalculator(TokenCalculator):
    encoding_name: str = Field(
        description="Encoding for converting text to tokens. "
        "Input encoding name or a specific OpenAI model",
        examples=["o200k_base", "gpt-4o"],
    )

    @field_validator("encoding_name")
    @classmethod
    def get_encoding_name(cls, value: str) -> str:
        try:
            enc = tiktoken.encoding_for_model(value)
            return enc.name
        except:
            try:
                tiktoken.get_encoding(value)
                return value
            except:
                return "o200k_base"

    def encode(self, text: str) -> list[int]:
        enc = tiktoken.get_encoding(self.encoding_name)
        return enc.encode(text=text)

    def calculate(self, text: str) -> int:
        enc_text = self.encode(text)
        return len(enc_text)

    def tokenize(
        self, text: str, decode_byte_str: bool = False, decoder: str = "utf-8"
    ):
        enc = tiktoken.get_encoding(self.encoding_name)
        enc_text = self.encode(text)
        tokens = [enc.decode_single_token_bytes(token_id) for token_id in enc_text]
        if decode_byte_str:
            tokens = [token.decode(decoder) for token in tokens]
        return tokens



from pydantic import BaseModel, ConfigDict


class OllamaEndpointRequestBody(BaseModel):
    model_config = ConfigDict(
        extra="forbid", use_enum_values=True, validate_assignment=True
    )


class OllamaEndpointResponseBody(BaseModel):
    model_config = ConfigDict(use_enum_values=True, validate_assignment=True)



from .add_upload_part import OpenAIUploadPartPathParam, OpenAIUploadPartRequestBody
from .cancel_upload import OpenAICancelUploadPathParam
from .complete_upload import (
    OpenAICompleteUploadPathParam,
    OpenAICompleteUploadRequestBody,
)
from .create_upload import OpenAIUploadRequestBody

__all__ = [
    "OpenAIUploadRequestBody",
    "OpenAIUploadPartPathParam",
    "OpenAIUploadPartRequestBody",
    "OpenAICompleteUploadPathParam",
    "OpenAICompleteUploadRequestBody",
    "OpenAICancelUploadPathParam",
]



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from lion.core.generic import LogManager, Pile, Progression
from lion.core.typing import ID, Any, BaseModel, JsonValue, Literal

from .action_request import ActionRequest
from .action_response import ActionResponse
from .assistant_response import AssistantResponse
from .instruction import Instruction
from .message import RoledMessage
from .system import System

DEFAULT_SYSTEM = "You are a helpful AI assistant. Let's think step by step."


class MessageManager:
    """
    Manages messages within a communication system.

    This class provides functionality for creating, adding, and managing
    different types of messages in a conversation. It maintains message
    history, handles system messages, and provides access to specific
    message types.

    Attributes:
        messages (Pile[RoledMessage]): Collection of messages
        logger (LogManager): Logger for message history
        system (System): System message setting context
        save_on_clear (bool): Whether to save logs when clearing
    """

    def __init__(self, messages=None, logger=None, system=None, save_on_clear=True):
        """
        Initialize a MessageManager instance.

        Args:
            messages: Initial list of messages
            logger: Logger instance for message history
            system: Initial system message
            save_on_clear: Whether to save logs when clearing
        """
        super().__init__()
        self.messages: Pile[RoledMessage] = Pile(
            items=messages, item_type={RoledMessage}
        )
        self.logger = logger or LogManager()
        self.system = system
        self.save_on_clear = save_on_clear
        if self.system:
            self.add_message(system=self.system)

    def set_system(self, system: System) -> None:
        """
        Set or update the system message.

        Args:
            system: The new system message to set
        """
        if not self.system:
            self.system = system
            self.messages.insert(0, self.system)
        else:
            old_system = self.system
            self.system = system
            self.messages.insert(0, self.system)
            self.messages.exclude(old_system)

    async def aclear_messages(self):
        """Asynchronously clear all messages except system message."""
        async with self.messages:
            self.clear_messages()

    async def a_add_message(self, **kwargs):
        """
        Asynchronously add a message.

        Args:
            **kwargs: Message creation parameters
        """
        async with self.messages:
            return self.add_message(**kwargs)

    @property
    def progress(self) -> Progression:
        """Get the progression of messages."""
        return Progression(order=list(self.messages))

    @staticmethod
    def create_instruction(
        *,
        sender: ID.SenderRecipient = None,
        recipient: ID.SenderRecipient = None,
        instruction: Instruction | JsonValue = None,
        context: JsonValue = None,
        guidance: JsonValue = None,
        plain_content: str = None,
        request_fields: list[str] | dict[str, Any] = None,
        request_model: type[BaseModel] | BaseModel = None,
        images: list = None,
        image_detail: Literal["low", "high", "auto"] = None,
        tool_schemas: dict | None = None,
        **kwargs,
    ) -> Instruction:
        """
        Create an instruction message.

        Args:
            sender: Message sender
            recipient: Message recipient
            instruction: Instruction content
            context: Additional context
            guidance: Optional guidance
            plain_content: Plain text content
            request_fields: Fields to request
            request_model: Pydantic model for requests
            images: Optional images
            image_detail: Image detail level
            tool_schemas: Optional tool schemas
            **kwargs: Additional parameters

        Returns:
            Instruction: The created instruction
        """
        if isinstance(instruction, Instruction):
            instruction.update(
                context,
                guidance=guidance,
                request_fields=request_fields,
                plain_content=plain_content,
                request_model=request_model,
                images=images,
                image_detail=image_detail,
                tool_schemas=tool_schemas,
                **kwargs,
            )
            if sender:
                instruction.sender = sender
            if recipient:
                instruction.recipient = recipient
            return instruction
        else:
            return Instruction(
                sender=sender,
                recipient=recipient,
                instruction=instruction,
                context=context,
                guidance=guidance,
                request_fields=request_fields,
                request_model=request_model,
                plain_content=plain_content,
                images=images,
                image_detail=image_detail,
                tool_schemas=tool_schemas,
                **kwargs,
            )

    @staticmethod
    def create_assistant_response(
        *,
        sender: Any = None,
        recipient: Any = None,
        assistant_response: AssistantResponse | Any = None,
    ) -> AssistantResponse:
        """
        Create an assistant response message.

        Args:
            sender: Message sender
            recipient: Message recipient
            assistant_response: Response content

        Returns:
            AssistantResponse: The created response
        """
        if isinstance(assistant_response, AssistantResponse):
            if sender:
                assistant_response.sender = sender
            if recipient:
                assistant_response.recipient = recipient
            return assistant_response

        return AssistantResponse(
            assistant_response=assistant_response,
            sender=sender,
            recipient=recipient,
        )

    @staticmethod
    def create_action_request(
        *,
        sender: ID.SenderRecipient = None,
        recipient: ID.SenderRecipient = None,
        function: str = None,
        arguments: dict[str, Any] = None,
        action_request: ActionRequest | None = None,
    ) -> ActionRequest:
        """
        Create an action request message.

        Args:
            sender: Message sender
            recipient: Message recipient
            function: Function to execute
            arguments: Function arguments
            action_request: Existing request to use

        Returns:
            ActionRequest: The created request

        Raises:
            ValueError: If action_request is not an ActionRequest instance
        """
        if action_request:
            if not isinstance(action_request, ActionRequest):
                raise ValueError(
                    "Error: action request must be an instance of ActionRequest."
                )
            if sender:
                action_request.sender = sender
            if recipient:
                action_request.recipient = recipient
            return action_request

        return ActionRequest(
            function=function,
            arguments=arguments,
            sender=sender,
            recipient=recipient,
        )

    @staticmethod
    def create_action_response(
        *,
        action_request: ActionRequest,
        action_response: ActionResponse | Any = None,
    ) -> ActionResponse:
        """
        Create an action response message.

        Args:
            action_request: The corresponding action request
            action_response: Response content

        Returns:
            ActionResponse: The created response

        Raises:
            ValueError: If action_request is invalid or already responded to
        """
        if not isinstance(action_request, ActionRequest):
            raise ValueError(
                "Error: please provide a corresponding action request for an "
                "action response."
            )

        if action_response:
            if isinstance(action_response, ActionResponse):
                if action_request.is_responded:
                    raise ValueError("Error: action request already has a response.")
                action_request.content["action_response_id"] = action_response.ln_id
                return action_response

        return ActionResponse(
            action_request=action_request,
            output=action_response,
        )

    @staticmethod
    def create_system(
        *,
        system: Any = None,
        sender: Any = None,
        recipient: Any = None,
        system_datetime: bool | str = None,
    ) -> System:
        """
        Create a system message.

        Args:
            system: System message content
            sender: Message sender
            recipient: Message recipient
            system_datetime: Whether to include datetime

        Returns:
            System: The created system message
        """
        system = system or DEFAULT_SYSTEM

        if isinstance(system, System):
            system.update(
                sender=sender,
                recipient=recipient,
                system_datetime=system_datetime,
            )
            return system

        return System(
            system=system,
            sender=sender,
            recipient=recipient,
            system_datetime=system_datetime,
        )

    def add_message(
        self,
        *,
        sender: ID.SenderRecipient = None,
        recipient: ID.SenderRecipient = None,
        instruction: Instruction | JsonValue = None,
        context: JsonValue = None,
        guidance: JsonValue = None,
        plain_content: str = None,
        request_fields: list[str] | dict[str, Any] = None,
        request_model: type[BaseModel] | BaseModel = None,
        images: list = None,
        image_detail: Literal["low", "high", "auto"] = None,
        assistant_response: AssistantResponse | Any = None,
        system: System | Any = None,
        system_datetime: bool | str = None,
        function: str = None,
        arguments: dict[str, Any] = None,
        action_request: ActionRequest | None = None,
        action_response: ActionResponse | Any = None,
        metadata: dict = None,
    ) -> RoledMessage:
        """
        Add a message to the manager.

        This method creates and adds a message of the specified type. Only
        one message type can be added at a time.

        Args:
            sender: Message sender
            recipient: Message recipient
            instruction: Instruction content
            context: Additional context
            guidance: Optional guidance
            plain_content: Plain text content
            request_fields: Fields to request
            request_model: Pydantic model for requests
            images: Optional images
            image_detail: Image detail level
            assistant_response: Assistant response content
            system: System message content
            system_datetime: Whether to include datetime
            function: Function for action request
            arguments: Arguments for action request
            action_request: Action request
            action_response: Action response
            metadata: Additional metadata

        Returns:
            RoledMessage: The added message

        Raises:
            ValueError: If multiple message types are specified
        """
        _msg = None
        if sum(bool(x) for x in (instruction, assistant_response, system)) > 1:
            raise ValueError("Only one message type can be added at a time.")

        if system:
            _msg = self.create_system(
                system=system,
                sender=sender,
                recipient=recipient,
                system_datetime=system_datetime,
            )
            self.set_system(_msg)

        elif action_response:
            if not action_request:
                raise ValueError("Error: Action response must have an action request.")
            _msg = self.create_action_response(
                action_request=action_request,
                action_response=action_response,
            )

        elif action_request or (function and arguments):
            _msg = self.create_action_request(
                sender=sender,
                recipient=recipient,
                function=function,
                arguments=arguments,
                action_request=action_request,
            )

        elif assistant_response:
            _msg = self.create_assistant_response(
                sender=sender,
                recipient=recipient,
                assistant_response=assistant_response,
            )

        else:
            _msg = self.create_instruction(
                sender=sender,
                recipient=recipient,
                instruction=instruction,
                context=context,
                guidance=guidance,
                plain_content=plain_content,
                request_fields=request_fields,
                request_model=request_model,
                images=images,
                image_detail=image_detail,
            )

        if metadata:
            _msg.metadata.update(["extra"], metadata)

        if _msg in self.messages:
            self.messages[_msg] = _msg
        else:
            self.messages.include(_msg)

        self.logger.log(_msg.to_log())
        return _msg

    def clear_messages(self) -> None:
        """Clear all messages except the system message."""
        if self.save_on_clear:
            self.logger.dump(clear=True)

        self.messages.clear()
        self.progress.clear()
        if self.system:
            self.messages.include(self.system)
            self.progress.insert(0, self.system)

    @property
    def last_response(self) -> AssistantResponse | None:
        """Get the last assistant response message."""
        for i in reversed(self.messages.progress):
            if isinstance(self.messages[i], AssistantResponse):
                return self.messages[i]

    @property
    def last_instruction(self) -> Instruction | None:
        """Get the last instruction message."""
        for i in reversed(self.messages.progress):
            if isinstance(self.messages[i], Instruction):
                return self.messages[i]

    @property
    def assistant_responses(self) -> Pile[AssistantResponse]:
        """Get all assistant response messages."""
        return Pile(
            [
                self.messages[i]
                for i in self.messages.progress
                if isinstance(self.messages[i], AssistantResponse)
            ]
        )

    @property
    def action_requests(self) -> Pile[ActionRequest]:
        """Get all action request messages."""
        return Pile(
            [
                self.messages[i]
                for i in self.messages.progress
                if isinstance(self.messages[i], ActionRequest)
            ]
        )

    @property
    def action_responses(self) -> Pile[ActionResponse]:
        """Get all action response messages."""
        return Pile(
            [
                self.messages[i]
                for i in self.messages.progress
                if isinstance(self.messages[i], ActionResponse)
            ]
        )

    @property
    def instructions(self) -> Pile[Instruction]:
        """Get all instruction messages."""
        return Pile(
            [
                self.messages[i]
                for i in self.messages.progress
                if isinstance(self.messages[i], Instruction)
            ]
        )

    def to_chat_msgs(self, progress=None) -> list[dict]:
        """
        Convert messages to chat format.

        Args:
            progress: Optional specific progression to convert

        Returns:
            list[dict]: Messages in chat format

        Raises:
            ValueError: If requested messages are not in the message pile
        """
        if progress == []:
            return []
        try:
            return [self.messages[i].chat_msg for i in (progress or self.progress)]
        except Exception as e:
            raise ValueError(
                "Invalid progress, not all requested messages are in the message pile"
            ) from e

    def __bool__(self):
        return not self.messages.is_empty()

    def has_logs(self):
        """Check if there are any logs."""
        return not self.logger.logs.is_empty()



from typing import Literal, Optional

from pydantic import Field

from ..data_models import OpenAIEndpointRequestBody
from .types import Endpoint


class OpenAIBatchRequestBody(OpenAIEndpointRequestBody):
    input_file_id: str = Field(
        description="The ID of an uploaded file that contains requests"
        " for the new batch."
    )

    endpoint: Endpoint = Field(
        description="The endpoint to be used for all requests in the batch."
    )

    completion_window: Literal["24h"] = Field(
        description="The time frame within which the batch" " should be processed."
    )

    metadata: dict | None = Field(
        None, description="Optional custom metadata for the batch."
    )



class Service:

    def list_tasks(self):
        pass


def register_service(cls):
    original_init = cls.__init__

    def wrapped_init(self, *args, **kwargs):
        original_init(self, *args, **kwargs)
        service_name = self.name
        ServiceSetting().add_service(self, service_name)

    cls.__init__ = wrapped_init
    return cls


class ServiceSetting:
    _instance = None
    services = {}

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def add_service(self, service: Service, name: str = None):
        if name:
            if self.services.get(name):
                raise ValueError(
                    "Invalid name. There is a service using the name, please change a name."
                )
            self.services[name] = service
        else:
            name = service.__class__.__name__ + "_" + str(len(self.services))
            self.services[name] = service



imported_models = {}


def match_response(request_model, response: dict | list):
    global imported_models

    endpoint = request_model.endpoint
    method = request_model.method

    # Messages endpoint
    if endpoint == "messages":
        if isinstance(response, dict):
            # Single message response
            if "AnthropicMessageResponseBody" not in imported_models:
                from .messages.response.response_body import (
                    AnthropicMessageResponseBody,
                )

                imported_models["AnthropicMessageResponseBody"] = (
                    AnthropicMessageResponseBody
                )

            # Convert to OpenAI-like format for AssistantResponse
            if response.get("content") and len(response["content"]) > 0:
                text_content = " ".join(
                    block["text"]
                    for block in response["content"]
                    if block["type"] == "text"
                )
                return {
                    "choices": [
                        {"message": {"content": text_content, "role": "assistant"}}
                    ],
                    "model": response.get("model"),
                    "usage": response.get("usage", {}),
                }
            return response
        else:
            # Stream response list
            if "AnthropicMessageResponseBody" not in imported_models:
                from .messages.response.response_body import (
                    AnthropicMessageResponseBody,
                )

                imported_models["AnthropicMessageResponseBody"] = (
                    AnthropicMessageResponseBody
                )

            # For streaming, we need to convert each chunk to OpenAI format
            result = []
            for item in response:
                if "type" not in item:
                    continue

                if item.get("type") == "content_block_delta":
                    if text := item.get("delta", {}).get(
                        "type"
                    ) == "text_delta" and item.get("delta", {}).get("text", {}):
                        result.append(
                            {
                                "choices": [
                                    {
                                        "delta": {
                                            "content": item["delta"]["text"],
                                            "role": "assistant",
                                        }
                                    }
                                ]
                            }
                        )
                elif item.get("type") == "message_delta":
                    # Include usage updates
                    if "usage" in item:
                        result.append(
                            {
                                "choices": [
                                    {"delta": {"content": "", "role": "assistant"}}
                                ],
                                "usage": item["usage"],
                            }
                        )
                elif item.get("type") == "message_stop":
                    # Final message with stop reason
                    result.append(
                        {
                            "choices": [
                                {
                                    "delta": {"content": "", "role": "assistant"},
                                    "finish_reason": "stop",
                                }
                            ],
                            "usage": item.get("usage", {}),
                        }
                    )

            return result

    raise ValueError(
        "There is no standard response model for the provided request and response"
    )



from typing import IO, List, Optional

from pydantic import BaseModel, ConfigDict, Field, field_validator

from ..data_models import OpenAIEndpointRequestBody, OpenAIEndpointResponseBody
from .types import TimestampGranularity, TranscriptionResponseFormat, WhisperModel


class OpenAITranscriptionRequestBody(OpenAIEndpointRequestBody):
    file: str | IO = Field(description="The audio file to transcribe")
    model: WhisperModel = Field(description="The model to use for transcription")
    language: str | None = Field(
        None, description="The language of the input audio (ISO-639-1 format)"
    )
    prompt: str | None = Field(
        None,
        description="An optional text to guide the model's style or"
        " continue a previous audio segment",
    )
    response_format: TranscriptionResponseFormat = Field(
        default="json",
        description="The format of the transcript output",
    )
    temperature: float = Field(
        default=0, ge=0, le=1, description="The sampling temperature"
    )
    timestamp_granularities: list[TimestampGranularity] | None = Field(
        default=["segment"],
        description="The timestamp granularities to populate " "for this transcription",
    )
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("file")
    @classmethod
    def get_file_object(cls, value):
        if isinstance(value, str):
            value = open(value, "rb")
        return value

    def __del__(self):
        # Ensure the file is closed when the object is destroyed
        file_obj = self.__dict__.get("file")
        if file_obj and not file_obj.closed:
            file_obj.close()


class OpenAITranscriptionResponseBody(OpenAIEndpointResponseBody):
    text: str = Field(description="The transcribed text")


class Word(BaseModel):
    word: str
    start: float
    end: float


class Segment(BaseModel):
    id: int
    seek: int
    start: float
    end: float
    text: str
    tokens: list[int]
    temperature: float
    avg_logprob: float
    compression_ratio: float
    no_speech_prob: float


class OpenAIVerboseTranscriptionResponseBody(OpenAIEndpointResponseBody):
    # TODO: check output object with different
    # timestamp_granularities in request body
    task: str = Field("transcribe", description="The task performed")
    language: str = Field(
        description="The detected or specified language of the input audio"
    )
    duration: float = Field(description="The duration of the input audio in seconds")
    text: str = Field(description="The transcribed text")
    words: list[Word] | None = Field(None, description="Word-level timestamps")
    segments: list[Segment] | None = Field(description="Segment-level details")

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "task": "transcribe",
                "language": "english",
                "duration": 8.470000267028809,
                "text": (
                    "The beach was a popular spot on a hot "
                    "summer day. People were swimming in the ocean,"
                    " building sandcastles, and playing beach volleyball."
                ),
                "segments": [
                    {
                        "id": 0,
                        "seek": 0,
                        "start": 0.0,
                        "end": 3.319999933242798,
                        "text": (
                            " The beach was a popular " "spot on a hot summer day."
                        ),
                        "tokens": [
                            50364,
                            440,
                            7534,
                            390,
                            257,
                            3743,
                            4008,
                            322,
                            257,
                            2368,
                            4266,
                            786,
                            13,
                            50530,
                        ],
                        "temperature": 0.0,
                        "avg_logprob": -0.2860786020755768,
                        "compression_ratio": 1.2363636493682861,
                        "no_speech_prob": 0.00985979475080967,
                    }
                ],
            }
        }
    )



def match_data_model(task_name):
    if task_name == "generate_completion":
        from .completion.request_body import OllamaCompletionRequestBody

        return {"request_body": OllamaCompletionRequestBody}

    elif task_name == "generate_chat_completion":
        from .chat_completion.request_body import OllamaChatCompletionRequestBody

        return {"request_body": OllamaChatCompletionRequestBody}

    elif task_name == "generate_embeddings":
        from .embedding.request_body import OllamaEmbeddingRequestBody

        return {"request_body": OllamaEmbeddingRequestBody}

    elif task_name == "create_model":
        from .model.create_model import OllamaCreateModelRequestBody

        return {"json_data": OllamaCreateModelRequestBody}

    elif task_name == "list_local_models" or task_name == "list_running_models":
        return {}

    elif task_name == "show_model_information":
        from .model.show_model import OllamaShowModelRequestBody

        return {"json_data": OllamaShowModelRequestBody}

    elif task_name == "copy_model":
        from .model.copy_model import OllamaCopyModelRequestBody

        return {"json_data": OllamaCopyModelRequestBody}

    elif task_name == "delete_model":
        from .model.delete_model import OllamaDeleteModelRequestBody

        return {"json_data": OllamaDeleteModelRequestBody}

    elif task_name == "pull_model":
        from .model.pull_model import OllamaPullModelRequestBody

        return {"json_data": OllamaPullModelRequestBody}

    elif task_name == "push_model":
        from .model.push_model import OllamaPushModelRequestBody

        return {"json_data": OllamaPushModelRequestBody}

    else:
        raise ValueError(f"Invalid task: {task_name}. Not supported in the service.")
